# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from hashlib import sha1
from platform import system

from ffpuppet import BrowserTerminatedError, BrowserTimeoutError, Debugger, Reason
from pytest import mark, raises

from sapphire import CertificateBundle

from .assets import AssetManager
from .puppet_target import PuppetTarget
from .target import Result, TargetLaunchError, TargetLaunchTimeout


def test_puppet_target_01(mocker, tmp_path):
    """test creating a PuppetTarget"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.reason = Reason.CLOSED
    fake_ffp.return_value.log_length.return_value = 562
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        assert target.asset_mgr
        assert target.closed
        assert target.launch_timeout == 300
        assert target.log_limit == 25
        assert target.memory_limit == 5000
        assert target.check_result(set()) == Result.NONE
        assert not target.https()
        assert target.log_size() == 1124
        fake_ffp.return_value.log_length.assert_any_call("stderr")
        fake_ffp.return_value.log_length.assert_any_call("stdout")
        assert target.monitor is not None
        target.save_logs(tmp_path / "fake_dest")
        assert fake_ffp.return_value.save_logs.call_count == 1
    assert fake_ffp.return_value.clean_up.call_count == 1
    # with extra args
    with PuppetTarget(fake_file, 1, 1, 1, rr=True, fake=1) as target:
        pass


def test_puppet_target_02(mocker, tmp_path):
    """test PuppetTarget.launch()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    # test providing prefs.js
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        # launch success
        target.launch("launch_target_page")
        assert fake_ffp.return_value.launch.call_count == 1
        assert fake_ffp.return_value.close.call_count == 0
        target.close()
        # launch timeout
        fake_ffp.reset_mock()
        fake_ffp.return_value.launch.side_effect = BrowserTimeoutError("timeout")
        with raises(TargetLaunchTimeout, match="timeout"):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.save_logs.call_count == 0
        # launch failure
        fake_ffp.reset_mock()
        (tmp_path / "log_stderr.txt").write_text("fake log")
        (tmp_path / "log_stdout.txt").write_text("fake log")
        mocker.patch(
            "grizzly.target.puppet_target.mkdtemp",
            autospec=True,
            return_value=str(tmp_path),
        )
        fake_ffp.return_value.launch.side_effect = BrowserTerminatedError("fail")
        with raises(TargetLaunchError, match="fail"):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.save_logs.call_count == 1


@mark.parametrize(
    "healthy, reason, ignore, result, closes",
    [
        # running as expected - no failures
        (True, None, [], Result.NONE, 0),
        # browser process closed
        (False, Reason.CLOSED, [], Result.NONE, 1),
        # browser process crashed
        (False, Reason.ALERT, [], Result.FOUND, 1),
        # browser exit with no crash logs
        (False, Reason.EXITED, [], Result.NONE, 1),
        # ffpuppet check failed
        (False, Reason.WORKER, [], Result.FOUND, 1),
        # ffpuppet check ignored (memory)
        (False, Reason.WORKER, ["memory"], Result.IGNORED, 1),
        # ffpuppet check ignored (log-limit)
        (False, Reason.WORKER, ["log-limit"], Result.IGNORED, 1),
    ],
)
def test_puppet_target_03(mocker, tmp_path, healthy, reason, ignore, result, closes):
    """test PuppetTarget.check_result()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    if "memory" in ignore:
        fake_ffp.return_value.available_logs.return_value = "ffp_worker_memory_usage"
    elif "log-limit" in ignore:
        fake_ffp.return_value.available_logs.return_value = "ffp_worker_log_size"
    fake_ffp.return_value.is_healthy.return_value = healthy
    fake_ffp.return_value.reason = reason
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        assert target.check_result(ignore) == result
    assert fake_ffp.return_value.close.call_count == closes


@mark.parametrize(
    "healthy, usage, os_name, killed, debugger",
    [
        # skip sending SIGABRT on unsupported OSs
        (True, [(1234, 90)], "Windows", 0, Debugger.NONE),
        # skip idle check if target is in a bad state
        (False, [], "Linux", 0, Debugger.NONE),
        # send SIGABRT to hung process
        (True, [(234, 10), (236, 75), (238, 60)], "Linux", 1, Debugger.NONE),
        # Don't send SIGABRT when using a debugger
        (True, [(236, 75)], "Linux", 0, Debugger.RR),
        # ignore idle timeout (close don't abort)
        (True, [(234, 10)], "Linux", 0, Debugger.NONE),
    ],
)
def test_puppet_target_04(mocker, tmp_path, healthy, usage, os_name, killed, debugger):
    """test PuppetTarget.handle_hang()"""
    mocker.patch(
        "grizzly.target.puppet_target.system", autospec=True, return_value=os_name
    )
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_kill = mocker.patch("grizzly.target.puppet_target.kill", autospec=True)
    # raise OSError for code coverage
    fake_kill.side_effect = OSError
    fake_file = tmp_path / "fake"
    fake_file.touch()
    fake_ffp.return_value.cpu_usage.return_value = usage
    fake_ffp.return_value.is_healthy.return_value = healthy
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target._debugger = debugger
        target.handle_hang()
    assert fake_ffp.return_value.is_healthy.call_count == 1
    assert fake_ffp.return_value.close.call_count == 1
    assert fake_ffp.return_value.cpu_usage.call_count == (1 if usage else 0)
    assert fake_kill.call_count == fake_ffp.return_value.wait.call_count == killed


@mark.skipif(system() != "Linux", reason="Linux only")
def test_puppet_target_05(mocker, tmp_path):
    """test PuppetTarget.dump_coverage()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.dump_coverage()
    assert fake_ffp.return_value.dump_coverage.call_count == 1


def test_puppet_target_06(mocker, tmp_path):
    """test PuppetTarget.monitor"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        fake_ffp.return_value.is_running.return_value = False
        fake_ffp.return_value.is_healthy.return_value = False
        assert target.monitor is not None
        assert not target.monitor.is_healthy()
        assert not target.monitor.is_running()
        fake_ffp.return_value.is_running.return_value = True
        fake_ffp.return_value.is_healthy.return_value = True
        assert target.monitor.is_healthy()
        assert target.monitor.is_running()
        fake_ffp.return_value.launches = 123
        assert target.monitor.launches == 123
        fake_ffp.return_value.log_length.return_value = 100
        assert target.monitor.log_length("stdout") == 100
        target.monitor.clone_log("somelog")
        assert fake_ffp.return_value.clone_log.call_count == 1


def test_puppet_target_07(mocker, tmp_path):
    """test PuppetTarget.monitor.is_idle()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.cpu_usage.return_value = [(999, 30), (998, 20), (997, 10)]
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        assert not target.monitor.is_idle(0)
        assert not target.monitor.is_idle(25)
        assert target.monitor.is_idle(50)


def test_puppet_target_08(mocker, tmp_path):
    """test PuppetTarget.process_assets()"""
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.write_text("1\n2\n")
    # no prefs file provided
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        assert target.asset_mgr.get("prefs") is None
        target.process_assets()
        asset = target.asset_mgr.get("prefs")
        assert asset
        assert asset.is_file()
        assert asset.name == "prefs.js"
    # prefs file provided
    with AssetManager(base_path=tmp_path) as asset_mgr:
        asset_mgr.add("prefs", fake_file)
        with PuppetTarget(fake_file, 300, 25, 5000) as target:
            target.asset_mgr = asset_mgr
            target.process_assets()
            asset = target.asset_mgr.get("prefs")
            assert asset
            assert asset.name == "fake"
    # abort tokens file provided
    with AssetManager(base_path=tmp_path) as asset_mgr:
        asset_mgr.add("abort-tokens", fake_file)
        with PuppetTarget(fake_file, 300, 25, 5000) as target:
            # ignore E1101: (pylint 2.9.3 bug?)
            #    Method 'add_abort_token' has no 'call_count' member (no-member)
            # pylint: disable=no-member
            assert target._puppet.add_abort_token.call_count == 0
            target.asset_mgr = asset_mgr
            target.process_assets()
            asset = target.asset_mgr.get("abort-tokens")
            assert asset
            assert asset.is_file()
            assert asset.name == "fake"
            assert target._puppet.add_abort_token.call_count == 2


@mark.parametrize(
    "pernosco, rr, valgrind",
    [
        # No debugger selected
        (False, False, False),
        # Pernosco selected
        (True, False, False),
        # rr selected
        (False, True, False),
        # Valgrind selected
        (False, False, True),
    ],  # pylint: disable=invalid-name
)
def test_puppet_target_09(
    mocker, tmp_path, pernosco, rr, valgrind
):  # pylint: disable=invalid-name
    """test PuppetTarget debugger args"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(
        fake_file, 30, 25, 500, pernosco=pernosco, rr=rr, valgrind=valgrind
    ) as _:
        pass
    if pernosco:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.PERNOSCO
    elif rr:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.RR
    elif valgrind:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.VALGRIND
    else:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.NONE


@mark.parametrize(
    "asset, env",
    [
        # suppressions via asset
        (True, False),
        # suppressions via env
        (False, True),
        # suppressions via both asset and env (asset should be preferred)
        (True, True),
        # missing suppressions file
        (False, False),
    ],
)
def test_puppet_target_10(tmp_path, asset, env):
    """test PuppetTarget.process_assets() - configure sanitizer suppressions"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    supp_asset = tmp_path / "supp_asset"
    supp_env = tmp_path / "supp_env"
    with AssetManager(base_path=tmp_path) as asset_mgr:
        asset_mgr.add("prefs", fake_file)
        if asset:
            supp_asset.touch()
            asset_mgr.add("lsan-suppressions", supp_asset)
        with PuppetTarget(fake_file, 300, 25, 5000) as target:
            target.environ["TSAN_OPTIONS"] = "a=1"
            if env:
                supp_env.touch()
                target.environ["LSAN_OPTIONS"] = f"suppressions='{supp_env}'"
            else:
                target.environ["LSAN_OPTIONS"] = "suppressions='missing'"
            target.asset_mgr = asset_mgr
            target.process_assets()
            if asset:
                assert (
                    f"suppressions='{target.asset_mgr.path / supp_asset.name}'"
                    in target.environ["LSAN_OPTIONS"]
                )
            elif env:
                assert (
                    f"suppressions='{target.asset_mgr.path / supp_env.name}'"
                    in target.environ["LSAN_OPTIONS"]
                )
            else:
                assert not asset_mgr.get("lsan-suppressions")


def test_puppet_target_11(mocker, tmp_path):
    """test PuppetTarget.filtered_environ()"""
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.environ = {
            "TRACKED": "1",
            "ASAN_OPTIONS": "external_symbolizer_path='a':no_remove='b'",
            "LSAN_OPTIONS": "suppressions='a'",
            "EMPTY": "",
        }
        assert "EMPTY" not in target.filtered_environ()
        assert "TRACKED" in target.filtered_environ()
        assert "ASAN_OPTIONS" in target.filtered_environ()
        assert "LSAN_OPTIONS" not in target.filtered_environ()


@mark.parametrize(
    "base, extra, result",
    [
        # empty
        ({}, {}, {}),
        # empty extra
        ({"a": "1"}, {}, {"a": "1"}),
        # empty base
        ({}, {"a": "1"}, {"a": "1"}),
        # merge
        ({"a": "1"}, {"b": "2"}, {"a": "1", "b": "2"}),
        # name collision, favor base
        ({"a": "1"}, {"a": "2"}, {"a": "1"}),
        # name collision and merge
        ({"a": "1"}, {"a": "2", "b": "2"}, {"a": "1", "b": "2"}),
    ],
)
def test_puppet_target_12(mocker, tmp_path, base, extra, result):
    """test PuppetTarget.merge_environment()"""
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.environ = base
        target.merge_environment(extra)
        assert target.environ == result


@mark.parametrize(
    "base, extra, result",
    [
        # collision with existing option
        ({"ASAN_OPTIONS": "a=1"}, {"ASAN_OPTIONS": "a=2"}, ["a=1"]),
        # add option from extra
        ({"ASAN_OPTIONS": "a=1"}, {"ASAN_OPTIONS": "b=2:c3"}, ["a=1", "b=2", "c=3"]),
        # add option from extra
        ({"ASAN_OPTIONS": "a=1:c=3"}, {"ASAN_OPTIONS": "b=2"}, ["a=1", "b=2", "c=3"]),
    ],
)
def test_puppet_target_13(mocker, tmp_path, base, extra, result):
    """test PuppetTarget.merge_environment() - merge sanitizer options"""
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.environ = base
        target.merge_environment(extra)
        for opt in target.environ["ASAN_OPTIONS"].split(":"):
            assert opt in result


def test_puppet_target_14(mocker, tmp_path):
    """test PuppetTarget.dump_coverage() - skip on unsupported platform"""
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    mocker.patch("grizzly.target.puppet_target.system", return_value="foo")
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.dump_coverage()


@mark.parametrize(
    "certutil, certs",
    [
        # certutil and cert bundle available
        (True, True),
        # missing certutil
        (False, True),
        # no cert bundle
        (True, False),
    ],
)
def test_puppet_target_15(mocker, tmp_path, certutil, certs):
    """test PuppetTarget - HTTPS support"""
    mocker.patch(
        "grizzly.target.puppet_target.certutil_available", return_value=certutil
    )
    mocker.patch("grizzly.target.puppet_target.certutil_find", autospec=True)
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)

    fake_file = tmp_path / "fake"
    fake_file.touch()

    bundle = mocker.Mock(spec_set=CertificateBundle).return_value if certs else None
    mocker.patch.object(PuppetTarget, "_get_certdb", return_value=tmp_path)
    with PuppetTarget(fake_file, 300, 25, 5000, certs=bundle) as target:
        assert target.https() == (certutil and certs)
        assert target._profile_template == (tmp_path if (certutil and certs) else None)


def test_puppet_target_16(mocker, tmp_path):
    """test PuppetTarget._get_certdb()"""
    (tmp_path / "grz_tmp").mkdir()
    mocker.patch(
        "grizzly.target.puppet_target.grz_tmp", return_value=tmp_path / "grz_tmp"
    )
    mocker.patch("grizzly.target.puppet_target.Profile", autospec=True)
    mocker.patch(
        "grizzly.target.puppet_target.find_cached", autospec=True, return_value=None
    )
    (tmp_path / "root.pem").touch()
    cert_hash = sha1(
        (tmp_path / "root.pem").read_bytes(), usedforsecurity=False
    ).hexdigest()
    certdb = tmp_path / f"certdb_{cert_hash}"
    certdb.mkdir()
    fake_add = mocker.patch(
        "grizzly.target.puppet_target.add_cached", autospec=True, return_value=tmp_path
    )
    PuppetTarget._get_certdb(tmp_path / "root.pem", "fake-certutil")
    assert fake_add.call_count == 1



"""
CertificateBundle unit tests
"""

# pylint: disable=protected-access

from pytest import raises

from .certificate_bundle import CertificateBundle


def test_certificate_bundle_basic(mocker, tmp_path):
    """test CertificateBundle"""
    mocker.patch("sapphire.certificate_bundle.mkdtemp", return_value=str(tmp_path))
    # create a bundle
    bundle = CertificateBundle.create()
    assert bundle.root.is_file()
    assert bundle.host.is_file()
    assert bundle.key.is_file()
    # load a bundle
    bundle = CertificateBundle.load(bundle._base)
    assert bundle.root.is_file()
    assert bundle.host.is_file()
    assert bundle.key.is_file()
    # remove bundle
    bundle.cleanup()
    assert not bundle._base.exists()
    # load empty path
    with raises(FileNotFoundError, match="does not exist"):
        CertificateBundle.load(tmp_path)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

from ..common.bugzilla import BugzillaBug
from ..common.frontend import Exit, configure_logging
from .args import ReplayFuzzBugzillaArgs
from .replay import ReplayManager

if TYPE_CHECKING:
    from argparse import Namespace

LOG = getLogger(__name__)


def main(args: Namespace | None = None) -> int:
    """CLI for `grizzly.replay.bugzilla`.

    Arguments:
        args: Result from `ReplayArgs.parse_args`.

    Returns:
        Exit.SUCCESS (0) for success otherwise a different Exit code is returned.
    """
    args = args or ReplayFuzzBugzillaArgs().parse_args()
    configure_logging(args.log_level)
    bug = BugzillaBug.load(args.input)
    if bug is None:
        LOG.info("Failed to load Bug %d from Bugzilla", args.input)
        return Exit.ERROR
    LOG.info("Loaded Bug %d from Bugzilla", args.input)
    with bug:
        args.asset.extend(
            # favor assets provided via the command line over the bug attachments
            bug.assets(ignore=tuple(x[0] for x in args.asset))
        )
        testcases = bug.testcases()
        if not testcases:
            LOG.error("No test case data attached to bug %d", args.input)
            return Exit.ERROR
        args.input = testcases
        return ReplayManager.main(args)


if __name__ == "__main__":
    raise SystemExit(main())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly Report"""
# pylint: disable=protected-access
from pathlib import Path

from FTB.Signatures.CrashInfo import CrashInfo
from pytest import mark

from .report import Report


def _create_crash_log(log_path):
    with log_path.open("w") as log_fp:
        log_fp.write("==1==ERROR: AddressSanitizer: SEGV on unknown address 0x0")
        log_fp.write(" (pc 0x0 bp 0x0 sp 0x0 T0)\n")
        log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write("    #1 0x1337dd in bar /file2.c:1806:19")


def test_report_01(tmp_path):
    """test Report() with boring logs (no stack)"""
    (tmp_path / "not_a_log.txt").touch()
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    report = Report(tmp_path, Path("a.bin"), size_limit=0)
    assert report._target_binary.name == "a.bin"
    assert report.path == tmp_path
    assert report._logs
    assert report._logs.aux is None
    assert report._logs.stderr.name == "log_stderr.txt"
    assert report._logs.stdout.name == "log_stdout.txt"
    assert report.preferred.name == "log_stderr.txt"
    assert report.stack is None
    assert report.major == Report.DEFAULT_MAJOR
    assert report.minor == Report.DEFAULT_MINOR
    assert report.prefix is not None
    report.cleanup()
    assert not tmp_path.exists()


def test_report_02(tmp_path):
    """test Report() with crash logs"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(tmp_path / "log_asan_blah.txt")
    report = Report(tmp_path, Path("bin"))
    assert report.path == tmp_path
    assert report._logs
    assert report._logs.aux.name == "log_asan_blah.txt"
    assert report._logs.stderr.name == "log_stderr.txt"
    assert report._logs.stdout.name == "log_stdout.txt"
    assert report.preferred.name == "log_asan_blah.txt"
    assert report.stack is not None
    assert report.major != Report.DEFAULT_MAJOR
    assert report.minor != Report.DEFAULT_MINOR
    assert report.prefix is not None
    report.cleanup()


def test_report_03(tmp_path):
    """test Report.tail()"""
    tmp_file = tmp_path / "file.txt"
    tmp_file.write_bytes(b"blah\ntest\n123\xEF\x00FOO")
    length = tmp_file.stat().st_size
    # don't trim
    Report.tail(tmp_file, length + 1)
    assert tmp_file.stat().st_size == length
    # perform trim
    Report.tail(tmp_file, 3)
    log_data = tmp_file.read_bytes()
    assert log_data.startswith(b"[LOG TAILED]\n")
    assert log_data[13:] == b"FOO"


def test_report_04(tmp_path):
    """test Report._select_logs() uninteresting data"""
    # test with empty path
    assert Report._select_logs(tmp_path) is None
    # empty file
    (tmp_path / "not_a_log.txt").touch()
    assert Report._select_logs(tmp_path) is None


def test_report_05(tmp_path):
    """test Report._select_logs()"""
    # small log with nothing interesting
    with (tmp_path / "log_asan.txt.1").open("wb") as log_fp:
        log_fp.write(b"SHORT LOG\n")
        log_fp.write(b"filler line")
    # crash on another thread
    with (tmp_path / "log_asan.txt.2").open("wb") as log_fp:
        log_fp.write(b"GOOD LOG\n")
        log_fp.write(
            b"==70811==ERROR: AddressSanitizer: SEGV on unknown address 0x00000BADF00D"
        )
        log_fp.write(
            b" (pc 0x7f4c0bb54c67 bp 0x7f4c07bea380 sp 0x7f4c07bea360 T0)\n"
        )  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    # child log that should be ignored (created when parent crashes)
    with (tmp_path / "log_asan.txt.3").open("wb") as log_fp:
        log_fp.write(b"BAD LOG\n")
        log_fp.write(
            b"==70811==ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000"
        )
        log_fp.write(
            b" (pc 0x7f4c0bb54c67 bp 0x7f4c07bea380 sp 0x7f4c07bea360 T2)\n"
        )  # must be 2nd line
        # pad out to 6 lines
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % l_no)
    (tmp_path / "log_mindump_blah.txt").write_bytes(b"minidump log")
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    # should be ignored in favor of "GOOD LOG"
    (tmp_path / "log_ffp_worker_blah.txt").write_bytes(b"worker log")
    log_map = Report._select_logs(tmp_path)
    assert log_map
    assert "GOOD LOG" in log_map.aux.read_text()
    assert "STDERR" in log_map.stderr.read_text()
    assert "STDOUT" in log_map.stdout.read_text()


def test_report_06(tmp_path):
    """test minidump with Report._select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_minidump_01.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|SIGSEGV|0x0|0\n")
        log_fp.write(b"minidump log\n")
    (tmp_path / "log_ffp_worker_blah.txt").write_bytes(b"worker log")
    log_map = Report._select_logs(tmp_path)
    assert log_map
    assert log_map.stderr.is_file()
    assert log_map.stdout.is_file()
    assert "minidump log" in log_map.aux.read_text()


def test_report_07(tmp_path):
    """test selecting preferred DUMP_REQUESTED minidump with Report._select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    with (tmp_path / "log_minidump_01.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f9518665d18|0\n")
        log_fp.write(b"0|0|bar.so|sadf|a.cc:739484451a63|3066|0x0\n")
        log_fp.write(b"0|1|gar.so|fdsa|b.cc:739484451a63|1644|0x12\n")
    with (tmp_path / "log_minidump_02.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f57ac9e2e14|0\n")
        log_fp.write(
            b"0|0|foo.so|google_breakpad::ExceptionHandler::WriteMinidump|"
            b"bar.cc:234|674|0xc\n"
        )
        log_fp.write(
            b"0|1|foo.so|google_breakpad::ExceptionHandler::WriteMinidump|"
            b"bar.cc:4a2|645|0x8\n"
        )
    with (tmp_path / "log_minidump_03.txt").open("wb") as log_fp:
        log_fp.write(b"GPU|||\n")
        log_fp.write(b"Crash|DUMP_REQUESTED|0x7f9518665d18|0\n")
        log_fp.write(b"0|0|bar.so|sadf|a.cc:1234|3066|0x0\n")
        log_fp.write(b"0|1|gar.so|fdsa|b.cc:4323|1644|0x12\n")
    log_map = Report._select_logs(tmp_path)
    assert log_map
    assert log_map.stderr.is_file()
    assert log_map.stdout.is_file()
    assert "google_breakpad::ExceptionHandler::WriteMinidump" in log_map.aux.read_text()


def test_report_08(tmp_path):
    """test selecting worker logs with Report._select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    (tmp_path / "log_ffp_worker_1.txt").write_bytes(b"worker log")
    # we should only ever see one but if we see multiple we warn, so test that.
    (tmp_path / "log_ffp_worker_2.txt").write_bytes(b"worker log")
    log_map = Report._select_logs(tmp_path)
    assert log_map
    assert log_map.stderr.is_file()
    assert log_map.stdout.is_file()
    assert "worker log" in log_map.aux.read_text()


def test_report_09(tmp_path):
    """test prioritizing sanitizer logs with Report._find_sanitizer()"""
    # NOTE: ordered by selection priority in order to use previously added logs
    # test empty
    (tmp_path / "log_asan.txt.0").touch()
    assert Report._find_sanitizer(list(tmp_path.iterdir())) is None
    # test *San log with data
    (tmp_path / "log_asan.txt.1").write_text("test")
    selected = Report._find_sanitizer(list(tmp_path.iterdir()))
    assert selected is not None
    assert "test" in selected.read_text()
    # test UBSan log
    (tmp_path / "log_asan.txt.1").write_text(
        "test.cc:3:5: runtime error: signed integer overflow: ..."
    )
    selected = Report._find_sanitizer(list(tmp_path.iterdir()))
    assert selected is not None
    assert "runtime error: signed integer overflow" in selected.read_text()
    # test selecting ASan report
    with (tmp_path / "log_asan.txt.2").open("wb") as log_fp:
        # missing stack
        log_fp.write(b"==1184==ERROR: AddressSanitizer: BUS on ... blah\n")
    with (tmp_path / "log_asan.txt.3").open("wb") as log_fp:
        log_fp.write(b"==9482==ERROR: AddressSanitizer: stack-overflow on ...\n")
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % (l_no,))
    selected = Report._find_sanitizer(list(tmp_path.iterdir()))
    assert selected is not None
    assert "AddressSanitizer: stack-overflow" in selected.read_text()
    # test selecting prioritized
    with (tmp_path / "log_asan.txt.4").open("wb") as log_fp:
        log_fp.write(
            b"==1942==ERROR: AddressSanitizer: heap-use-after-free on ... blah\n"
        )
        for l_no in range(4):
            log_fp.write(b"    #%d blah...\n" % (l_no,))
    with (tmp_path / "log_asan.txt.5").open("wb") as log_fp:
        log_fp.write(b"==1984==ERROR: AddressSanitizer: SEGV on ... blah\n")
        log_fp.write(b"missing trace...\n")
    with (tmp_path / "log_asan.txt.6").open("wb") as log_fp:
        log_fp.write(b"ERROR: Failed to mmap\n")
    selected = Report._find_sanitizer(list(tmp_path.iterdir()))
    assert selected is not None
    assert "heap-use-after-free" in selected.read_text()
    # test selecting TSan reports
    tsan_path = tmp_path / "tsan"
    tsan_path.mkdir()
    (tsan_path / "log_asan_benign.txt").write_text(
        "==27531==WARNING: Symbolizer buffer too small\n"
        "==27531==WARNING: Symbolizer buffer too small"
    )
    tsan_report = tsan_path / "log_asan_report.txt"
    tsan_report.write_text(
        "WARNING: ThreadSanitizer: data race (pid=26919)\n"
        "  Write of size 8 at 0x7f0ca2fc3400 by thread T51:\n"
        "    #0 memcpy /sanitizer_common_interceptors.inc:810:5 (lib+0x6656e)\n"
    )
    selected = Report._find_sanitizer(list(tsan_path.iterdir()))
    assert selected is not None
    assert selected == tsan_report


def test_report_10(tmp_path):
    """test Report() size_limit"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log\n" * 200)
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log\n" * 200)
    (tmp_path / "unrelated.txt").write_bytes(b"nothing burger\n" * 200)
    (tmp_path / "rr-trace").mkdir()
    size_limit = len("STDERR log\n")
    report = Report(tmp_path, Path("bin"), size_limit=size_limit)
    assert report.path == tmp_path
    assert report._logs
    assert report._logs.aux is None
    assert report._logs.stderr.name == "log_stderr.txt"
    assert report._logs.stdout.name == "log_stdout.txt"
    assert report.preferred.name == "log_stderr.txt"
    assert report.stack is None
    size_limit += len("[LOG TAILED]\n")
    assert report._logs.stderr.stat().st_size == size_limit
    assert report._logs.stdout.stat().st_size == size_limit
    assert (tmp_path / "unrelated.txt").stat().st_size == size_limit
    report.cleanup()
    assert not tmp_path.is_dir()


def test_report_11(tmp_path):
    """test selecting Valgrind logs with Report._select_logs()"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    (tmp_path / "log_valgrind.txt").write_bytes(b"valgrind log")
    log_map = Report._select_logs(tmp_path)
    assert log_map
    assert log_map.stderr.is_file()
    assert log_map.stdout.is_file()
    assert "valgrind log" in log_map.aux.read_text()


def test_report_12(tmp_path):
    """test Report.crash_info"""
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(tmp_path / "log_asan_blah.txt")
    # no binary.fuzzmanagerconf
    report = Report(tmp_path, Path("bin"))
    assert report._crash_info is None
    assert report.crash_info is not None
    assert report._crash_info is not None
    assert report._crash_info.configuration.product == "bin"
    # with binary.fuzzmanagerconf
    with (tmp_path / "bin.fuzzmanagerconf").open("wb") as conf:
        conf.write(b"[Main]\n")
        conf.write(b"platform = x86-64\n")
        conf.write(b"product = grizzly-test\n")
        conf.write(b"os = linux\n")
    report = Report(tmp_path, Path("bin"))
    report._target_binary = tmp_path / "bin"
    assert report._crash_info is None
    assert report.crash_info is not None
    assert report._crash_info is not None
    assert report._crash_info.configuration.product == "grizzly-test"


@mark.parametrize(
    "sig_cache, has_sig",
    [
        # signature exists in cache
        ('{"symptoms": [{"functionNames": ["a"],"type": "stackFrames"}]}', True),
        # no signature
        (None, True),
        # FM failed to generate signature
        (None, False),
    ],
)
def test_report_13(mocker, tmp_path, sig_cache, has_sig):
    """test Report.crash_signature and Report.crash_hash"""
    mocker.patch("grizzly.common.report.ProgramConfiguration", autospec=True)
    collector = mocker.patch("grizzly.common.report.Collector", autospec=True)
    if sig_cache:
        sig_file = tmp_path / "cache.sig"
        sig_file.write_text(sig_cache)
        collector.return_value.search.return_value = (str(sig_file), None)
        collector.return_value.sigCacheDir = str(tmp_path)
    else:
        collector.return_value.sigCacheDir = None
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    if has_sig:
        _create_crash_log(tmp_path / "log_asan_blah.txt")
    report = Report(tmp_path, Path("bin"))
    assert report._signature is None
    if has_sig:
        assert report.crash_signature
        assert report.crash_info.createShortSignature() == "[@ foo]"
    else:
        assert not report.crash_signature
    assert report.crash_hash


@mark.parametrize(
    "backtrace, lines",
    [
        # baseline
        ((), 5),
        # not ignored
        (("1",), 5),
        # ignore frame
        (("std::panicking::rust_panic",), 6),
        # ignore frame
        (("alloc::alloc",), 6),
        # ignore frame and other frames
        (("1", "std::panicking::rust_panic", "3"), 6),
        # ignored frames is last ignorable frame
        (("1", "2", "3", "4", "std::panicking::rust_panic", "6"), 6),
        # ignored frames is too deep
        (("1", "2", "3", "4", "5", "std::panicking::rust_panic"), 5),
    ],
)
def test_report_14(mocker, backtrace, lines):
    """test Report.crash_signature_max_frames()"""
    info = mocker.Mock(spec=CrashInfo, backtrace=backtrace)
    assert Report.crash_signature_max_frames(info, suggested_frames=5) == lines


@mark.parametrize(
    "data, lines",
    [
        # simple log
        ("test", 1),
        # simple log
        ("a\nb", 2),
        # empty log
        ("", 0),
        # log with bad chars
        ("a\n\0", 2),
    ],
)
def test_report_15(tmp_path, data, lines):
    """test Report._load_log()"""
    log = tmp_path / "test-log.txt"
    log.write_text(data)
    assert len(Report._load_log(log)) == lines


@mark.parametrize(
    "hang, has_log, expected",
    [
        # process log and create short signature
        (False, True, "[@ foo]"),
        # no log available to create short signature
        (False, False, "Signature creation failed"),
        # result is a hang
        (True, True, "Potential hang detected"),
    ],
)
def test_report_16(mocker, tmp_path, hang, has_log, expected):
    """test Report.short_signature"""
    mocker.patch("grizzly.common.report.ProgramConfiguration", autospec=True)
    collector = mocker.patch("grizzly.common.report.Collector", autospec=True)
    # prevent checking signature cache (if it exists)
    collector.return_value.sigCacheDir = None
    (tmp_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (tmp_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    if has_log:
        _create_crash_log(tmp_path / "log_asan_blah.txt")
    report = Report(tmp_path, Path("bin"), is_hang=hang)
    assert report.short_signature == expected
    if hang:
        assert report.crash_hash == "hang"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""Unit tests for `grizzly.reduce.strategies`."""
from collections import namedtuple
from logging import getLogger
from pathlib import Path

from pytest import mark, raises

from sapphire import Sapphire

from ..common.report import Report
from ..common.storage import TestCase
from ..replay import ReplayResult
from ..target import AssetManager, Target
from . import ReduceManager
from .strategies import Strategy, _load_strategies

LOG = getLogger(__name__)
pytestmark = mark.usefixtures(
    "reporter_sequential_strftime",
    "tmp_path_status_db_reduce",
)


@mark.parametrize("is_hang", [True, False])
def test_strategy_tc_load(tmp_path, is_hang):
    """test that strategy base class dump and load doesn't change testcase metadata"""

    class _TestStrategy(Strategy):
        def __iter__(self):
            yield [TestCase.load(x) for x in sorted(self._testcase_root.iterdir())]

        def update(self, success):
            pass

    # create testcase
    with TestCase("a.htm", "adpt", input_fname="fn") as test:
        test.duration = 1.2
        test.hang = is_hang
        test.add_from_bytes(b"123", test.entry_point)
        test.dump(tmp_path / "src", include_details=True)

    with _TestStrategy([TestCase.load(tmp_path / "src")]) as strategy:
        for attempt in strategy:
            assert len(attempt) == 1
            assert attempt[0].hang == is_hang
            strategy.update(False)


def _fake_save_logs_foo(result_logs):
    """write fake log data to disk"""
    (result_logs / "log_stderr.txt").write_text("STDERR log\n")
    (result_logs / "log_stdout.txt").write_text("STDOUT log\n")
    (result_logs / "log_asan_blah.txt").write_text(
        "==1==ERROR: AddressSanitizer: "
        "SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n"
        "    #0 0xbad000 in foo /file1.c:123:234\n"
        "    #1 0x1337dd in bar /file2.c:1806:19\n"
    )


def test_strategy_load_fail(mocker):
    """test that a broken strategy doesn't block other strategies"""

    class _GoodStrategy(Strategy):
        name = "good"

        @classmethod
        def load(cls):
            return cls

    class _BadStrategy:
        name = "bad"

        @classmethod
        def load(cls):
            raise RuntimeError("oops")

    mocker.patch(
        "grizzly.reduce.strategies.iter_entry_points",
        return_value=(_BadStrategy, _GoodStrategy),
    )
    mocker.patch("grizzly.reduce.strategies.DEFAULT_STRATEGIES", new=("good",))
    result = _load_strategies()
    assert result == {"good": _GoodStrategy}

    # missing strategy in defaults should raise
    mocker.patch("grizzly.reduce.strategies.DEFAULT_STRATEGIES", new=("good", "bad"))
    with raises(AssertionError):
        _load_strategies()


ListStrategyParams = namedtuple(
    "ListStrategyParams",
    "test_data, strategies, required_first, expected_run_calls, expected_results,"
    "expected_num_reports",
)


@mark.parametrize(
    ListStrategyParams._fields,
    [
        # "list" is a no-op with a single testcase
        ListStrategyParams(
            test_data=[b"123"],
            strategies=["check", "list"],
            required_first=False,
            expected_run_calls=1,
            expected_results={"123"},
            expected_num_reports=2,
        ),
        # "list" reduces a list of testcases down to one required
        ListStrategyParams(
            test_data=[b"", b"", b"123", b""],
            strategies=["list"],
            required_first=False,
            expected_run_calls=4,
            expected_results={"123"},
            expected_num_reports=2,
        ),
        # "list" reduces a list of testcases down to two required
        ListStrategyParams(
            test_data=[b"", b"required", b"123", b""],
            strategies=["list"],
            required_first=True,
            expected_run_calls=4,
            expected_results={"required", "123"},
            expected_num_reports=3,
        ),
    ],
)
def test_list(
    mocker,
    tmp_path,
    test_data,
    strategies,
    required_first,
    expected_run_calls,
    expected_results,
    expected_num_reports,
):
    """tests for the "list" strategy"""
    mocker.patch("grizzly.reduce.strategies.lithium._contains_dd", return_value=True)
    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value

    def replay_run(testcases, _time_limit, **kw):
        kw["on_iteration_cb"]()
        required_seen = False
        for test in testcases:
            contents = test["test.html"].read_text()
            LOG.debug("interesting if %r == '123'", contents)
            if contents == "required" and required_first:
                required_seen = True
            elif contents == "123" and (required_seen or not required_first):
                log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
                log_path.mkdir()
                _fake_save_logs_foo(log_path)
                report = Report(log_path, Path("bin"))
                return [ReplayResult(report, [], True)]
        return []

    replayer.run.side_effect = replay_run

    tests = []
    for num, data in enumerate(test_data):
        with TestCase("test.html", "test-adapter") as test:
            test.add_from_bytes(data, test.entry_point)
            test.dump(tmp_path / "src" / f"{num:02d}", include_details=True)
        tests.append(TestCase.load(tmp_path / "src" / f"{num:02d}"))
    log_path = tmp_path / "logs"

    target = mocker.Mock(spec_set=Target)
    target.filtered_environ.return_value = {}
    target.asset_mgr = mocker.Mock(spec_set=AssetManager)
    with ReduceManager(
        set(),
        mocker.Mock(spec_set=Sapphire, timeout=30),
        target,
        tests,
        strategies,
        log_path,
        use_analysis=False,
    ) as mgr:
        assert mgr.run() == 0

    assert replayer.run.call_count == expected_run_calls
    assert set(log_path.iterdir()) == {log_path / "reports"}
    tests = {test.read_text() for test in log_path.glob("reports/*-*/test.html")}
    assert tests == expected_results
    assert (
        sum(1 for _ in (log_path / "reports").iterdir()) == expected_num_reports
    ), list((log_path / "reports").iterdir())


def test_dd_only(mocker, tmp_path):
    """test that only files containing DDBEGIN/END are reduced"""
    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value

    def replay_run(testcases, _time_limit, **kw):
        kw["on_iteration_cb"]()
        for test in testcases:
            contents = test["test.html"].read_text()
            LOG.debug("interesting if 'required' in %r", contents)
            interesting = "required" in contents
            if interesting:
                log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
                log_path.mkdir()
                _fake_save_logs_foo(log_path)
                report = Report(log_path, Path("bin"))
                return [ReplayResult(report, [], True)]
        return []

    replayer.run.side_effect = replay_run

    with TestCase("test.html", "test-adapter") as test:
        test.add_from_bytes(b"DDBEGIN\n123\nrequired\nDDEND\n", "test.html")
        test.add_from_bytes(b"blah\n", "other.html")
        test.dump(tmp_path / "src", include_details=True)
    test = TestCase.load(tmp_path / "src", catalog=True)
    tests = [test]
    log_path = tmp_path / "logs"

    target = mocker.Mock(spec_set=Target)
    target.filtered_environ.return_value = {}
    target.asset_mgr = mocker.Mock(spec_set=AssetManager)
    with ReduceManager(
        set(),
        mocker.Mock(spec_set=Sapphire, timeout=30),
        target,
        tests,
        ["lines"],
        log_path,
        use_analysis=False,
    ) as mgr:
        assert mgr.run() == 0

    expected_run_calls = 3
    expected_results = {"DDBEGIN\nrequired\nDDEND\n"}
    expected_num_reports = 2

    assert replayer.run.call_count == expected_run_calls
    assert set(log_path.iterdir()) == {log_path / "reports"}
    assert len(tuple(log_path.glob("reports/*-*/*.html"))) == 2
    tests = {test.read_text() for test in log_path.glob("reports/*-*/test.html")}
    assert tests == expected_results
    assert (
        sum(1 for _ in (log_path / "reports").iterdir()) == expected_num_reports
    ), list((log_path / "reports").iterdir())
    others = {test.read_text() for test in log_path.glob("reports/*-*/other.html")}
    assert others == {"blah\n"}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from base64 import b64encode
from pathlib import Path
from zipfile import ZipFile

from bugsy import Attachment, Bug, BugsyException
from pytest import fixture, mark
from requests.exceptions import ConnectionError as RequestsConnectionError

from .bugzilla import BugzillaBug


@fixture(autouse=True)
def tmp_path_grz_tmp(tmp_path, mocker):
    """Provide an alternate working directory for testing."""

    def _grz_tmp(*subdir):
        path = Path(tmp_path, "grizzly", *subdir)
        path.mkdir(parents=True, exist_ok=True)
        return path

    mocker.patch("grizzly.common.bugzilla.grz_tmp", _grz_tmp)


def test_bugzilla_01(mocker):
    """test BugzillaBug._fetch_attachments()"""
    bug = mocker.Mock(spec=Bug, id=123)
    bug.get_attachments.return_value = (
        # ignored obsolete
        mocker.Mock(spec=Attachment, is_obsolete=True),
        # ignored content type
        mocker.Mock(
            spec=Attachment,
            is_obsolete=False,
            content_type="text/x-phabricator-request",
        ),
        # ignored file extension
        mocker.Mock(
            spec=Attachment,
            is_obsolete=False,
            content_type="text/plain",
            file_name="ignore.txt",
        ),
        # valid test case
        mocker.Mock(
            spec=Attachment,
            is_obsolete=False,
            content_type="text/html",
            file_name="test.html",
            data=b64encode(b"foo"),
        ),
        # corrupted data
        mocker.Mock(
            spec=Attachment,
            is_obsolete=False,
            content_type="text/html",
            file_name="broken.html",
            data=b"bad-b64",
        ),
        # invalid file name
        mocker.Mock(
            spec=Attachment,
            is_obsolete=False,
            content_type="text/html",
            file_name=".",
            data=b64encode(b"foo"),
        ),
        # invalid file name
        mocker.Mock(
            spec=Attachment,
            is_obsolete=False,
            content_type="text/html",
            file_name="../escape.html",
            data=b64encode(b"foo"),
        ),
    )
    with BugzillaBug(bug) as bz_bug:
        assert len(tuple(bz_bug.path.iterdir())) == 1
        assert (bz_bug.path / "test.html").is_file()
        assert (bz_bug.path / "test.html").read_text() == "foo"
        assert not (bz_bug.path / ".." / "escape.html").is_file()


@mark.parametrize(
    "exc",
    [
        BugsyException("foo", error_code=101),
        BugsyException("access denied", error_code=102),
        RequestsConnectionError(),
    ],
)
def test_bugzilla_02(mocker, exc):
    """test BugzillaBug.load() - errors"""
    bugsy = mocker.patch("grizzly.common.bugzilla.Bugsy", autospec=True)
    bugsy.return_value.get.side_effect = exc
    bugsy.return_value.bugzilla_url = "foo"
    assert BugzillaBug.load(123) is None


def test_bugzilla_03(mocker):
    """test BugzillaBug.load()"""
    bugsy = mocker.patch("grizzly.common.bugzilla.Bugsy", autospec=True)
    bugsy.return_value.get.return_value = mocker.MagicMock(spec=Bug, id=123)
    with BugzillaBug.load(123):
        pass


def test_bugzilla_04(mocker):
    """test BugzillaBug.assets()"""
    bug = mocker.Mock(spec=Bug, id=123)
    bug.get_attachments.return_value = []
    with BugzillaBug(bug) as bz_bug:
        (bz_bug._data / "prefs.js").touch()
        (bz_bug._data / "test.html").touch()
        # load prefs.js asset
        results = tuple(bz_bug.assets())
        assert len(results) == 1
        assert results[0] == ("prefs", bz_bug._data / "prefs.js")
        # load ignore prefs.js asset
        assert not any(bz_bug.assets(ignore=["prefs"]))


@mark.parametrize("archive_count", [0, 1, 2])
def test_bugzilla_05(mocker, tmp_path, archive_count):
    """test BugzillaBug._unpack_archive()"""
    (tmp_path / "test.html").write_text("foo")
    bug = mocker.Mock(spec=Bug, id=123)
    bug.get_attachments.return_value = []
    with BugzillaBug(bug) as bz_bug:
        (bz_bug._data / "not_archive.txt").touch()
        for num in range(archive_count):
            with ZipFile(bz_bug._data / f"archive{num:02d}.zip", "w") as zfp:
                zfp.write(tmp_path / "test.html", arcname="test.html")
        bz_bug._unpack_archives()
        results = tuple(x for x in bz_bug._data.iterdir() if x.is_dir())
        assert len(results) == archive_count
        for num in range(archive_count):
            assert not (bz_bug._data / f"archive{num:02d}.zip").is_file()
        for path in results:
            assert (path / "test.html").is_file()


@mark.parametrize(
    "files, count",
    [
        # no files
        ((), 0),
        # single test case
        (("test.html",), 1),
        # multiple test cases (add the path and both files individually)
        (("a.html", "b.html"), 3),
        # test case and asset
        (("prefs.js", "test.html"), 1),
        # test loading archive
        (("archive.zip",), 1),
        # test loading archive and standalone test
        (("archive.zip", "test.html"), 2),
    ],
)
def test_bugzilla_06(mocker, tmp_path, files, count):
    """test BugzillaBug.testcases()"""
    (tmp_path / "test.html").write_text("foo")
    bug = mocker.Mock(spec=Bug, id=123)
    bug.get_attachments.return_value = []
    with BugzillaBug(bug) as bz_bug:
        for file in files:
            if file.endswith("zip"):
                with ZipFile(bz_bug._data / file, "w") as zfp:
                    zfp.write(tmp_path / "test.html", arcname="test.html")
            else:
                (bz_bug._data / file).touch()
        assert len(bz_bug.testcases()) == count



"""
Job unit tests
"""

# pylint: disable=protected-access
from pathlib import Path
from platform import system

from pytest import mark, raises

from .job import Job, Served
from .server_map import (
    DynamicResource,
    FileResource,
    IncludeResource,
    RedirectResource,
    ServerMap,
)


def test_job_01(tmp_path):
    """test creating a simple Job"""
    with raises(RuntimeError, match="Empty Job"):
        Job(tmp_path)
    test_file = tmp_path / "test.txt"
    test_file.touch()
    job = Job(tmp_path, required_files=[test_file.name])
    assert not job.forever
    assert job.status == Served.NONE
    assert job.lookup_resource("") is None
    assert job.lookup_resource("test") is None
    assert job.lookup_resource("test/test/") is None
    assert job.lookup_resource("test/../../") is None
    assert job.lookup_resource("\x00\x0B\xAD\xF0\x0D") is None
    assert job.lookup_resource("test.txt")
    assert job.pending == 1
    assert not job.is_complete()
    assert job.remove_pending(str(test_file))
    assert job.pending == 0
    job.finish()
    assert not any(job.served)
    assert job.is_complete(wait=0.01)


def test_job_02(tmp_path):
    """test Job proper handling of required and optional files"""
    opt = []
    opt.append(tmp_path / "opt_file_1.txt")
    opt[-1].write_bytes(b"a")
    req = []
    req.append(tmp_path / "req_file_1.txt")
    req[-1].write_bytes(b"b")
    (tmp_path / "nested").mkdir()
    opt.append(tmp_path / "nested" / "opt_file_2.txt")
    opt[-1].write_bytes(b"c")
    req.append(tmp_path / "nested" / "req_file_2.txt")
    req[-1].write_bytes(b"d")
    job = Job(tmp_path, required_files=[req[0].name, f"nested/{req[1].name}"])
    assert job.status == Served.NONE
    assert not job.is_complete()
    resource = job.lookup_resource("req_file_1.txt")
    assert isinstance(resource, FileResource)
    assert resource.required
    assert job.pending == 2
    assert resource.target == tmp_path / "req_file_1.txt"
    assert not job.remove_pending("no_file.test")
    assert job.pending == 2
    assert not job.remove_pending(str(req[0]))
    job.mark_served(resource)
    assert len(job._served.files) == 1
    assert job.status == Served.REQUEST
    assert job.pending == 1
    resource = job.lookup_resource("nested/req_file_2.txt")
    assert job.remove_pending(str(req[1]))
    job.mark_served(resource)
    assert len(job._served.files) == 2
    assert job.status == Served.ALL
    assert job.pending == 0
    assert job.remove_pending(str(req[0]))
    job.mark_served(resource)
    assert len(job._served.files) == 2
    resource = job.lookup_resource("opt_file_1.txt")
    assert isinstance(resource, FileResource)
    assert not resource.required
    assert resource.target == opt[0]
    assert job.remove_pending(str(opt[0]))
    job.mark_served(resource)
    assert len(job._served.files) == 3
    assert len(job.served) == 3
    resource = job.lookup_resource("nested/opt_file_2.txt")
    assert isinstance(resource, FileResource)
    assert resource.target == opt[1]
    assert not resource.required
    job.finish()
    assert job.is_complete()


def test_job_03(tmp_path):
    """test Job redirects"""
    smap = ServerMap()
    smap.set_redirect("one", "somefile.txt", required=False)
    smap.set_redirect("two", "reqfile.txt")
    job = Job(tmp_path, server_map=smap)
    assert job.status == Served.NONE
    resource = job.lookup_resource("one")
    assert isinstance(resource, RedirectResource)
    resource = job.lookup_resource("two")
    assert isinstance(resource, RedirectResource)
    assert job.pending == 1
    assert job.remove_pending("two")
    assert job.pending == 0


def test_job_04(mocker, tmp_path):
    """test Job includes"""
    srv_root = tmp_path / "root"
    srv_include = tmp_path / "test"
    srv_include_2 = tmp_path / "test_2"
    srv_include_nested = srv_include / "nested"
    srv_root.mkdir()
    srv_include.mkdir()
    srv_include_2.mkdir()
    srv_include_nested.mkdir()
    test_1 = srv_root / "req_file.txt"
    test_1.write_bytes(b"a")
    inc_1 = srv_include / "test_file.txt"
    inc_1.write_bytes(b"b")
    nst_1 = srv_include_nested / "nested_file.txt"
    nst_1.write_bytes(b"c")
    inc_2 = srv_include_2 / "test_file_2.txt"
    inc_2.write_bytes(b"d")
    # stub out ServerMap._check_url() because it is too
    # restrictive to allow testing of some functionality
    mocker.patch.object(ServerMap, "_check_url", side_effect=lambda x: x)
    smap = ServerMap()
    smap.set_include("testinc", srv_include)
    # add manually to avoid sanity checks in ServerMap.set_include()
    smap.include["testinc/fakedir"] = IncludeResource(
        "testinc/fakedir", False, srv_include
    )
    smap.include["testinc/1/2/3"] = IncludeResource("testinc/1/2/3", False, srv_include)
    smap.include[""] = IncludeResource("", False, srv_include)
    smap.set_include("testinc/inc2", srv_include_2)
    job = Job(srv_root, server_map=smap, required_files=[test_1.name])
    assert job.status == Served.NONE
    # test include path pointing to a missing file
    assert job.lookup_resource("testinc/missing") is None
    # test include path pointing to an invalid file name
    assert job.lookup_resource("testinc/\x00\x0B\xAD\xF0\x0D") is None
    # test includes that map to 'srv_include'
    includes = [res for res in smap.include.values() if res.target == srv_include]
    assert includes
    for include in includes:
        request = f"{include.url}/test_file.txt"
        resource = job.lookup_resource(request)
        assert isinstance(resource, FileResource)
        assert srv_include in resource.target.parents
        assert resource.url == request.lstrip("/")
    # test nested include path pointing to a different include
    request = "testinc/inc2/test_file_2.txt"
    resource = job.lookup_resource(request)
    assert isinstance(resource, FileResource)
    assert resource.target == inc_2
    assert resource.url == request
    # test redirect root without leading '/'
    request = "test_file.txt"
    resource = job.lookup_resource(request)
    assert isinstance(resource, FileResource)
    assert resource.target == srv_include / "test_file.txt"
    assert resource.url == request
    # test redirect with file in a nested directory
    request = "testinc/nested/nested_file.txt"
    resource = job.lookup_resource(request)
    assert isinstance(resource, FileResource)
    assert resource.target == nst_1
    assert resource.url == request


def test_job_05(tmp_path):
    """test Job.lookup_resource() with tricky includes"""
    srv_root = tmp_path / "root"
    srv_root.mkdir()
    req = srv_root / "req_file.txt"
    req.write_bytes(b"a")
    inc_dir = tmp_path / "inc_dir"
    inc_dir.mkdir()
    (inc_dir / "sub").mkdir()
    inc_file1 = inc_dir / "sub" / "include.js"
    inc_file1.write_bytes(b"a")
    inc_file2 = inc_dir / "test_inc.html"
    inc_file2.write_bytes(b"a")
    # test url matching part of the file name
    smap = ServerMap()
    smap.include["inc_url"] = IncludeResource("inc_url", False, inc_dir)
    job = Job(srv_root, server_map=smap, required_files=[req.name])
    resource = job.lookup_resource("inc_url/sub/include.js")
    assert isinstance(resource, FileResource)
    assert resource.target == inc_file1
    # test checking only the include url
    assert job.lookup_resource("inc_url") is None
    # file and include file collision (files should always win)
    smap.include.clear()
    inc_a = inc_dir / "a.bin"
    inc_a.write_bytes(b"a")
    file_a = srv_root / "a.bin"
    file_a.write_bytes(b"a")
    smap.include["/"] = IncludeResource("", False, inc_dir)
    resource = job.lookup_resource("a.bin")
    assert isinstance(resource, FileResource)
    assert resource.target == file_a
    # TODO: inc and inc subdir collision can fail.
    # /inc/a file
    # /inc   a/file
    # Should we detect and support it? Does it matter?


def test_job_06(tmp_path):
    """test Job dynamic"""
    smap = ServerMap()
    smap.set_dynamic_response("cb1", lambda _: b"0", mime_type="a/a", required=True)
    smap.set_dynamic_response("cb2", lambda _: b"1")
    job = Job(tmp_path, server_map=smap)
    assert job.status == Served.NONE
    assert job.pending == 1
    resource = job.lookup_resource("cb1")
    assert isinstance(resource, DynamicResource)
    assert callable(resource.target)
    assert isinstance(resource.mime, str)
    resource = job.lookup_resource("cb2")
    assert isinstance(resource, DynamicResource)
    assert callable(resource.target)
    assert isinstance(resource.mime, str)


def test_job_07(tmp_path):
    """test accessing forbidden files"""
    srv_root = tmp_path / "root"
    srv_root.mkdir()
    test_file = srv_root / "req_file.txt"
    test_file.write_bytes(b"a")
    no_access = tmp_path / "no_access.txt"
    no_access.write_bytes(b"a")
    job = Job(srv_root, required_files=[test_file.name])
    assert job.status == Served.NONE
    assert job.pending == 1
    assert job.lookup_resource("../no_access.txt") is None


@mark.skipif(system() == "Windows", reason="Unsupported on Windows")
def test_job_08(tmp_path):
    """test Job with file names containing invalid characters"""
    test_file = tmp_path / "test.txt"
    test_file.write_bytes(b"a")
    (tmp_path / "?_2.txt").write_bytes(b"a")
    job = Job(tmp_path, required_files=[test_file.name])
    assert job.status == Served.NONE
    assert job.pending == 1
    resource = job.lookup_resource("test.txt")
    assert resource
    assert resource.target == test_file


def test_job_09(tmp_path):
    """test Job.lookup_resource() with file name that is too long"""
    (tmp_path / "test.txt").touch()
    job = Job(tmp_path, required_files=["test.txt"])
    assert job.status == Served.NONE
    assert job.pending == 1
    assert job.lookup_resource(f"/{'a' * 8192}.txt") is None


def test_job_10(tmp_path):
    """test Job with missing directory"""
    with raises(OSError):
        Job(tmp_path / "missing")


def test_job_11(tmp_path):
    """test Job.mark_served() and Job.served"""
    (tmp_path / "test.txt").touch()
    job = Job(tmp_path, required_files=["test.txt"])
    assert not any(job.served)
    # add first resource
    resource = FileResource("a.bin", False, tmp_path / "a.bin", "mine/mime")
    job.mark_served(resource)
    assert "a.bin" in job.served
    assert job.served[resource.url] == resource.target
    assert len(job.served) == 1
    # add a resource with the same url
    job.mark_served(FileResource("a.bin", False, tmp_path / "a.bin", "mine/mime"))
    assert len(job._served.files) == 1
    assert len(job.served) == 1
    # add a nested resource
    resource = FileResource("nested/b.bin", False, tmp_path / "nested" / "b.bin", "a/a")
    job.mark_served(resource)
    assert "nested/b.bin" in job.served
    assert job.served[resource.url] == resource.target
    assert len(job.served) == 2
    # add an include resource
    resource = IncludeResource("inc.bin", False, Path("/some/include/path/inc.bin"))
    job.mark_served(resource)
    assert "inc.bin" in job.served
    assert job.served[resource.url] == resource.target
    assert len(job.served) == 3
    # add an include resource pointing to a common file with unique url
    resource = IncludeResource("alt_path", False, Path("/some/include/path/inc.bin"))
    job.mark_served(resource)
    assert "alt_path" in job.served
    assert len(job.served) == 4
    assert job.served[resource.url] == resource.target


def test_job_12():
    """test Job.lookup_mime()"""
    assert Job.lookup_mime("unknown") == "application/octet-stream"
    # look up from Job.MIME_MAP
    assert ".avif" in Job.MIME_MAP, "test is broken"
    assert Job.lookup_mime("test.avif") == "image/avif"
    # look up known ext
    assert Job.lookup_mime("test.html") == "text/html"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ..replay.bucket import bucket_main
from .args import ReduceFuzzManagerIDQualityArgs
from .crash import main as crash_main


def main() -> int:
    """Wrapper for bucket_main() which is the CLI for `grizzly.reduce.bucket`.

    Arguments:
        None

    Returns:
        Exit.SUCCESS (0) for success otherwise a different Exit code is returned.
    """
    return bucket_main(ReduceFuzzManagerIDQualityArgs().parse_args(), crash_main)


if __name__ == "__main__":
    raise SystemExit(main())



"""
ConnectionManager unit tests
"""

# pylint: disable=protected-access

from itertools import count
from socket import socket

from pytest import mark, raises

from .connection_manager import ConnectionManager
from .job import Job
from .worker import Worker


@mark.parametrize("timeout", [10, 0])
def test_connection_manager_01(mocker, tmp_path, timeout):
    """test basic ConnectionManager"""
    (tmp_path / "testfile").write_bytes(b"test")
    job = Job(tmp_path, required_files=["testfile"])
    clnt_sock = mocker.Mock(spec_set=socket)
    clnt_sock.recv.return_value = b"GET /testfile HTTP/1.1"
    serv_sock = mocker.Mock(spec_set=socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    mocker.patch("sapphire.worker.select", return_value=([serv_sock], None, None))
    assert not job.is_complete()
    with ConnectionManager(job, serv_sock) as mgr:
        assert mgr.serve(timeout)
    assert clnt_sock.close.call_count == 1
    assert job.is_complete()
    assert not job.accepting.is_set()
    assert job.exceptions.empty()


@mark.parametrize("worker_limit", [1, 2, 10])
def test_connection_manager_02(mocker, tmp_path, worker_limit):
    """test ConnectionManager multiple files and requests"""
    (tmp_path / "test1").touch()
    (tmp_path / "test2").touch()
    (tmp_path / "test3").touch()
    job = Job(tmp_path, required_files=["test1", "test2", "test3"])
    clnt_sock = mocker.Mock(spec_set=socket)
    clnt_sock.recv.side_effect = (
        b"GET /test1 HTTP/1.1",
        b"GET /missing HTTP/1.1",
        b"badrequest",
        b"",
        b"GET /test2 HTTP/1.1",
        b"GET /test1 HTTP/1.1",
        b"GET /test1 HTTP/1.1",
        b"GET /test3 HTTP/1.1",
    )
    serv_sock = mocker.Mock(spec_set=socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    mocker.patch("sapphire.worker.select", return_value=([serv_sock], None, None))
    assert not job.is_complete()
    with ConnectionManager(job, serv_sock, limit=worker_limit) as mgr:
        assert mgr.serve(10)
    assert clnt_sock.close.call_count == 8
    assert job.is_complete()


def test_connection_manager_03(mocker, tmp_path):
    """test ConnectionManager re-raise worker exceptions"""
    (tmp_path / "file").touch()
    job = Job(tmp_path, required_files=["file"])
    clnt_sock = mocker.Mock(spec_set=socket)
    clnt_sock.recv.side_effect = Exception("worker exception")
    serv_sock = mocker.Mock(spec_set=socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    mocker.patch("sapphire.worker.select", return_value=([serv_sock], None, None))
    with (
        raises(Exception, match="worker exception"),
        ConnectionManager(job, serv_sock) as mgr,
    ):
        mgr.serve(10)
    assert clnt_sock.close.call_count == 1
    assert job.is_complete()
    assert job.exceptions.empty()


def test_connection_manager_04(mocker, tmp_path):
    """test ConnectionManager.serve() with callback"""
    (tmp_path / "file").touch()
    job = Job(tmp_path, required_files=["file"])
    with ConnectionManager(job, mocker.Mock(spec_set=socket), poll=0.01) as mgr:
        # invalid callback
        with raises(TypeError, match="continue_cb must be callable"):
            mgr.serve(10, continue_cb="test")
        # job did not start
        assert not job.is_complete()
        # callback abort
        assert mgr.serve(10, continue_cb=lambda: False)
        assert job.is_complete()


def test_connection_manager_05(mocker, tmp_path):
    """test ConnectionManager.serve() with timeout"""
    mocker.patch("sapphire.connection_manager.perf_counter", side_effect=count())
    (tmp_path / "file").touch()
    clnt_sock = mocker.Mock(spec_set=socket)
    clnt_sock.recv.return_value = b""
    serv_sock = mocker.Mock(spec_set=socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    mocker.patch("sapphire.worker.select", return_value=([serv_sock], None, None))
    job = Job(tmp_path, required_files=["file"])
    with ConnectionManager(job, serv_sock, poll=0.01) as mgr:
        assert not mgr.serve(10)
        assert job.is_complete()


def test_connection_manager_06(mocker, tmp_path):
    """test ConnectionManager.serve() worker fails to exit"""
    mocker.patch("sapphire.worker.Thread", autospec=True)
    mocker.patch("sapphire.connection_manager.perf_counter", side_effect=count())
    (tmp_path / "file").touch()
    clnt_sock = mocker.Mock(spec_set=socket)
    serv_sock = mocker.Mock(spec_set=socket)
    serv_sock.accept.return_value = (clnt_sock, None)
    mocker.patch("sapphire.worker.select", return_value=([serv_sock], None, None))
    job = Job(tmp_path, required_files=["file"])
    mocker.patch.object(job, "worker_complete")
    with ConnectionManager(job, serv_sock) as mgr:
        with raises(RuntimeError, match="Failed to close workers"):
            mgr.serve(10)
        assert job.is_complete()
    assert clnt_sock.close.call_count == 1


def test_connection_manager_07(mocker):
    """test ConnectionManager._join_workers()"""
    # no workers
    assert not ConnectionManager._join_workers([])
    # worker fails to join, without timeout
    fake_worker = mocker.Mock(spec_set=Worker)
    fake_worker.join.return_value = False
    assert ConnectionManager._join_workers([fake_worker], timeout=0)
    assert fake_worker.join.call_count == 1
    fake_worker.reset_mock()
    # worker fails to join, with timeout
    assert ConnectionManager._join_workers([fake_worker], timeout=1)
    assert fake_worker.join.call_count == 1
    fake_worker.reset_mock()
    # worker joins
    fake_worker.join.return_value = True
    assert not ConnectionManager._join_workers([fake_worker], timeout=0)
    assert fake_worker.join.call_count == 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

__all__ = ("TargetMonitor",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


class TargetMonitor(metaclass=ABCMeta):
    @abstractmethod
    def clone_log(self, log_id: str, offset: int = 0) -> Path | None:
        """Create a copy of a log.

        Args:
            log_id: Log identifier.
            offset: Number of bytes to seek into log before copying data.

        Returns:
            Copy of specified log.
        """

    @abstractmethod
    def is_healthy(self) -> bool:
        """Check for failures such as assertions, crashes, etc.

        Args:
            None

        Returns:
            True if no target failures are found otherwise False.
        """

    @abstractmethod
    def is_idle(self, threshold: int) -> bool:
        """Check if target is idle.

        Args:
            threshold: Maximum allowed CPU usage as percentage (per process).

        Returns:
            True if CPU usage for all processes is below the threshold otherwise False.
        """

    @abstractmethod
    def is_running(self) -> bool:
        """Check if target is running.

        Args:
            None

        Returns:
            True if target is running otherwise False.
        """

    @property
    @abstractmethod
    def launches(self) -> int:
        """Number of successful target launches.

        Args:
            None

        Returns:
            Number of successful launches.
        """

    @abstractmethod
    def log_length(self, log_id: str) -> int:
        """Calculate the length of a specific log file.

        Args:
            log_id: Log identifier.

        Returns:
            Log file size in bytes.
        """



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

from ..common.frontend import Exit, configure_logging
from ..common.fuzzmanager import load_fm_data
from ..common.reporter import Quality
from ..replay.crash import modify_args
from .args import ReduceFuzzManagerIDArgs
from .core import ReduceManager

if TYPE_CHECKING:
    from argparse import Namespace

LOG = getLogger(__name__)


def main(args: Namespace | None = None) -> int:
    """CLI for `grizzly.reduce.crash`.

    Arguments:
        args: Result from `ReduceArgs.parse_args`.

    Returns:
        Exit.SUCCESS (0) for success otherwise a different Exit code is returned.
    """
    args = args or ReduceFuzzManagerIDArgs().parse_args()
    configure_logging(args.log_level)
    with load_fm_data(args.input, load_bucket=not args.sig) as (crash, bucket):
        LOG.info(
            "Loaded crash %d (%s) from FuzzManager",
            crash.crash_id,
            Quality(crash.testcase_quality).name,
        )
        # call grizzly.reduce
        result = ReduceManager.main(modify_args(args, crash, bucket))

        # update quality
        # map Exit.* -> Quality.*
        # default back to UNREDUCED
        # most errors will not be related to the testcase
        # so they should be retried later
        if args.fuzzmanager:
            quality = {
                Exit.ERROR: Quality.REDUCER_ERROR,
                Exit.ABORT: Quality(crash.testcase_quality),
                Exit.SUCCESS: Quality.ORIGINAL,
                Exit.FAILURE: Quality(args.no_repro_quality),
            }.get(Exit(result), Quality.UNREDUCED)
            # don't ever set things back to REDUCING, default to UNREDUCED in that case.
            # REDUCING is only used in automation, so ABORT should never happen.
            if quality == Quality.REDUCING:
                quality = Quality.UNREDUCED
            LOG.info(
                "reducer finished -> exit(%d) -> %s (Q%d)",
                result,
                quality.name,
                quality,
            )
            crash.testcase_quality = quality.value

    return result


if __name__ == "__main__":
    raise SystemExit(main())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Common unit test fixtures for `grizzly.replay`.
"""

from pathlib import Path

from pytest import fixture

from sapphire import Sapphire


@fixture
def server(mocker):
    """Mock Sapphire server"""
    srv = mocker.Mock(spec_set=Sapphire, port=1337, timeout=10)
    srv_cls = mocker.patch("grizzly.replay.replay.Sapphire", autospec=True)
    srv_cls.return_value.__enter__.return_value = srv
    return srv


@fixture
def tmp_path_grz_tmp(tmp_path, mocker):
    """Provide an alternate working directory for testing."""

    def _grz_tmp(*subdir):
        path = Path(tmp_path, "grizzly", *subdir)
        path.mkdir(parents=True, exist_ok=True)
        return path

    mocker.patch("grizzly.replay.replay.grz_tmp", _grz_tmp)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from importlib.metadata import PackageNotFoundError

from pytest import mark

from .utils import grz_tmp, package_version


def test_grz_tmp_01(mocker, tmp_path):
    """test grz_tmp()"""
    fake_tmp = tmp_path / "grizzly"
    mocker.patch("grizzly.common.utils.GRZ_TMP", new=fake_tmp)
    # create temp path
    assert not fake_tmp.is_dir()
    path = grz_tmp()
    assert path == fake_tmp
    assert path.is_dir()
    # create temp path (exists)
    assert grz_tmp() == fake_tmp
    # create temp path with sub directory
    path = grz_tmp("test1", "test2")
    assert path == fake_tmp / "test1" / "test2"
    assert path.is_dir()


@mark.parametrize(
    "version, expected",
    [
        # missing package
        (PackageNotFoundError(), "unknown"),
        # success
        (("1.2.3",), "1.2.3"),
    ],
)
def test_package_version_01(mocker, version, expected):
    """test package_version()"""
    mocker.patch("grizzly.common.utils.version", autospec=True, side_effect=version)
    assert package_version("foo", default="unknown") == expected



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly main"""
from platform import system

from pytest import mark, skip

from sapphire import CertificateBundle

from .args import GrizzlyArgs
from .common.frontend import Exit
from .main import main
from .target import TargetLaunchError


@mark.parametrize(
    "adpt_relaunch, extra_args",
    [
        # successful run
        (0, []),
        # successful run - no harness
        (0, ["--no-harness"]),
        # successful run (with iteration limit)
        (0, ["--limit", "10"]),
        # successful run (with runtime limit)
        (0, ["--runtime", "10"]),
        # successful run (with coverage)
        (0, ["--coverage"]),
        # adapter relaunch 1
        (1, []),
        # relaunch 10
        (0, ["--relaunch", "10"]),
        # smoke test detects result
        (0, ["--smoke-test"]),
        # ignore something
        (0, ["--ignore", "timeout"]),
        # non-default display
        (0, ["--display", "headless"]),
        # verbose mode
        (0, ["--verbose"]),
    ],
)
def test_main_01(mocker, session_setup, adpt_relaunch, extra_args):
    """test main()"""
    mocker.patch("grizzly.main.get_certs", autospec=True)
    cov_support = system() == "Linux"
    if "--coverage" in extra_args and not cov_support:
        skip(f"--coverage not available on {system()}")
    mocker.patch(
        "grizzly.args.scan_plugins",
        autospec=True,
        side_effect=(["targ"], ["adpt"]),
    )
    adapter, session_cls, session_obj, target_cls = session_setup
    adapter.RELAUNCH = adpt_relaunch

    # use __file__ as "binary" since it is not used
    cmd = [__file__, "adpt", "--platform", "targ", *extra_args]
    args = GrizzlyArgs().parse_args(cmd)
    session_obj.status.results.total = 1 if args.smoke_test else 0

    assert main(args) == (Exit.ERROR if args.smoke_test else Exit.SUCCESS)
    if cov_support:
        assert session_cls.mock_calls[0][-1]["coverage"] == args.coverage
    if adpt_relaunch:
        assert session_cls.mock_calls[0][-1]["relaunch"] == adpt_relaunch
    else:
        assert session_cls.mock_calls[0][-1]["relaunch"] == args.relaunch
    assert session_obj.run.call_count == 1
    assert target_cls.return_value.cleanup.call_count == 1


@mark.parametrize(
    "exit_code, to_raise",
    [
        # test user abort
        (Exit.ABORT, KeyboardInterrupt),
        # test launch failure
        (Exit.LAUNCH_FAILURE, TargetLaunchError),
    ],
)
def test_main_02(mocker, session_setup, exit_code, to_raise):
    """test main() - exit codes"""
    mocker.patch("grizzly.main.FailedLaunchReporter", autospec=True)
    _, _, session_obj, target_cls = session_setup
    if to_raise == TargetLaunchError:
        session_obj.run.side_effect = TargetLaunchError("test", mocker.Mock())
    else:
        session_obj.run.side_effect = to_raise()
    args = mocker.MagicMock(adapter="fake", time_limit=1, timeout=1)
    assert main(args) == exit_code
    assert target_cls.return_value.cleanup.call_count == 1


@mark.parametrize(
    "test_limit, timeout",
    [
        # use default test time limit and timeout values
        (None, None),
        # set test time limit
        (10, None),
        # set both test time limit and timeout to the same value
        (10, 10),
        # set timeout greater than test time limit
        (10, 11),
        # use default test time limit and low timeout
        (None, 1),
    ],
)
def test_main_03(mocker, session_setup, test_limit, timeout):
    """test main() - time-limit and timeout"""
    # no_harness=False for code coverage
    args = mocker.MagicMock(
        adapter="fake",
        no_harness=False,
        time_limit=test_limit,
        timeout=timeout,
    )
    assert main(args) == Exit.SUCCESS
    target_cls = session_setup[3]
    assert target_cls.return_value.cleanup.call_count == 1


@mark.parametrize(
    "pernosco, rr, valgrind",
    [
        # No debugger enabled
        (False, False, False),
        # Pernosco enabled
        (True, False, False),
        # rr enabled
        (False, True, False),
        # Valgrind enabled
        (False, False, True),
    ],
)  # pylint: disable=invalid-name
def test_main_04(
    mocker, session_setup, pernosco, rr, valgrind
):  # pylint: disable=invalid-name
    """test enabling debuggers"""
    # maximum one debugger allowed at a time
    assert sum((pernosco, rr, valgrind)) < 2, "test broken!"
    args = mocker.MagicMock(
        adapter="fake",
        pernosco=pernosco,
        rr=rr,
        time_limit=1,
        timeout=1,
        valgrind=valgrind,
    )
    assert main(args) == Exit.SUCCESS
    target_cls = session_setup[3]
    assert target_cls.call_args[-1]["pernosco"] == pernosco
    assert target_cls.call_args[-1]["rr"] == rr
    assert target_cls.call_args[-1]["valgrind"] == valgrind


def test_main_05(mocker, tmp_path, session_setup):
    """test target does not support https"""
    (tmp_path / "bundle").mkdir()
    bundle = CertificateBundle.create(tmp_path / "bundle")
    mocker.patch("grizzly.main.get_certs", autospec=True, return_value=bundle)
    target_cls = session_setup[3]
    target_cls.return_value.https.return_value = False
    args = mocker.MagicMock(
        adapter="fake",
        use_http=False,
        time_limit=1,
        timeout=1,
    )
    assert main(args) == Exit.SUCCESS
    assert target_cls.return_value.https.call_count == 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import TYPE_CHECKING

from ..adapter import Adapter

if TYPE_CHECKING:
    from sapphire import ServerMap

    from ...common.storage import TestCase

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class NoOpAdapter(Adapter):
    """This is a simple adapter that has very little overhead. It can be used
    to help measure a baseline iteration rate for Grizzly.
    """

    NAME = "no-op"

    def setup(self, input_path: str | None, server_map: ServerMap) -> None:
        """Generate a static test case that calls `window.close()` when run.
        Normally this is done in generate() but since the test is static only
        do it once. Use the default harness to allow running multiple test cases
        in a row without closing the browser after each one.

        Args:
            input: Unused in this example.
            server_map: Unused in this example.

        Returns:
            None
        """
        self.enable_harness()
        self.fuzz["test"] = (
            b"<!DOCTYPE html>\n"
            b"<html>\n"
            b"<head>\n"
            b"<script>window.close()</script>\n"
            b"</head>\n"
            b"</html>"
        )

    def generate(self, testcase: TestCase, server_map: ServerMap) -> None:
        """The test case contents have been created now add the data to the TestCase.

        All TestCases require an entry point and the one expected by Grizzly
        is provided in `testcase.entry_point` so use it as the file name for
        the test.

        Args:
            testcase: TestCase to be populated.
            server_map: Unused in this example.

        Returns:
            None
        """
        testcase.add_from_bytes(self.fuzz["test"], testcase.entry_point, required=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
unit tests for grizzly.replay.args
"""
from platform import system

from pytest import mark, param, raises

from .args import ReplayArgs, ReplayFuzzManagerIDArgs, ReplayFuzzManagerIDQualityArgs


def test_replay_args_01(capsys, mocker, tmp_path):
    """test parsing args"""
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=["targ1"])
    # missing args tests
    with raises(SystemExit):
        ReplayArgs().parse_args([])
    # specified prefs.js missing
    exe = tmp_path / "binary"
    exe.touch()
    # test success
    ReplayArgs().parse_args([str(exe), str(exe), "--platform", "targ1"])
    # test missing input
    with raises(SystemExit):
        ReplayArgs().parse_args(argv=[str(exe), "missing", "--platform", "targ1"])
    assert "error: 'missing' does not exist" in capsys.readouterr()[-1]


@mark.parametrize(
    "args, msg, inputs",
    [
        # test any-crash with signature
        (
            ["--any-crash", "--sig", "x"],
            "error: signature is ignored when running with --any-crash",
            1,
        ),
        # test in valid idle_delay
        (
            ["--idle-threshold", "1", "--idle-delay", "-1"],
            "error: --idle-delay value must be positive",
            1,
        ),
        # test invalid min-crashes value
        (["--min-crashes", "0"], "error: --min-crashes value must be positive", 1),
        # test invalid repeat value
        (["--repeat", "-1"], "error: --repeat value must be positive", 1),
        # test running with rr without --output set
        param(
            ["--rr"],
            "error: --output must be set when using rr",
            1,
            marks=[mark.skipif(system() != "Linux", reason="Linux only")],
        ),
        # test missing signature file
        (["--sig", "missing"], "error: signature file not found", 1),
        # test --no-harness with multiple inputs
        (
            ["--no-harness"],
            "error: '--no-harness' cannot be used with multiple testcases",
            3,
        ),
    ],
)
def test_replay_args_02(capsys, mocker, tmp_path, args, msg, inputs):
    """test ReplayArgs.parse_args() - sanity checks"""
    mocker.patch("grizzly.args.Path.read_text", autospec=True, return_value="0")
    target = "target1"
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=[target])
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    with raises(SystemExit):
        ReplayArgs().parse_args(
            argv=[str(fake_bin)]
            + [str(fake_bin) for _ in range(inputs)]
            + ["--platform", target]
            + args
        )
    assert msg in capsys.readouterr()[-1]


def test_replay_args_03(tmp_path, capsys):
    """test ReplayFuzzManagerIDArgs"""
    exe = tmp_path / "binary"
    exe.touch()
    ReplayFuzzManagerIDArgs().parse_args([str(exe), "123"])
    # error cases
    with raises(SystemExit):
        ReplayFuzzManagerIDArgs().parse_args(
            [str(exe), "123", "--no-harness", "--test-index", "0", "1"]
        )
    assert (
        "error: '--test-index' only supports a single value with '--no-harness'"
        in capsys.readouterr()[-1]
    )


def test_replay_args_04(capsys, tmp_path):
    """test ReplayFuzzManagerIDQualityArgs"""
    exe = tmp_path / "binary"
    exe.touch()
    with raises(SystemExit):
        ReplayFuzzManagerIDQualityArgs().parse_args(
            [str(exe), "123", "--quality", "-1"]
        )
    assert "error: '--quality' value cannot be negative" in capsys.readouterr()[-1]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from enum import IntEnum, unique
from logging import DEBUG, basicConfig, getLogger
from os import getenv, getpid
from pathlib import Path
from shutil import copytree
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING

from sapphire import CertificateBundle

from .cache import add_cached, find_cached
from .utils import grz_tmp

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .storage import TestCase

__all__ = (
    "DEFAULT_TIME_LIMIT",
    "TIMEOUT_DELAY",
    "ConfigError",
    "Exit",
    "configure_logging",
    "display_time_limits",
    "time_limits",
)

DEFAULT_TIME_LIMIT = 30
LOG = getLogger(__name__)
# TIMEOUT_DELAY is added to the test time limit to create the default timeout
TIMEOUT_DELAY = 15


class ConfigError(Exception):
    """Raised to indicate invalid configuration a state"""

    def __init__(self, message: str, exit_code: int) -> None:
        super().__init__(message)
        self.exit_code = exit_code


@unique
class Exit(IntEnum):
    """Exit codes"""

    SUCCESS = 0
    # unexpected error occurred (invalid input, unhanded exception, etc)
    ERROR = 1
    # invalid argument
    ARGS = 2
    # run aborted (ctrl+c, etc)
    ABORT = 3
    # unrelated Target failure (browser startup crash, etc)
    LAUNCH_FAILURE = 4
    # expected results not reproduced (opposite of SUCCESS)
    FAILURE = 5


def configure_logging(log_level: int) -> None:
    """Configure log output level and formatting.

    Args:
        log_level: Set log level.

    Returns:
        None
    """
    # allow force enabling log_level via environment
    if getenv("DEBUG", "0").lower() in ("1", "true"):
        log_level = DEBUG
    if log_level == DEBUG:
        date_fmt = None
        log_fmt = "%(asctime)s %(levelname).1s %(name)s | %(message)s"
    else:
        date_fmt = "%Y-%m-%d %H:%M:%S"
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt=date_fmt, level=log_level)


def display_time_limits(time_limit: int, timeout: int, no_harness: bool) -> None:
    """Output configuration of time limits and harness.

    Args:
        time_limit: Time in seconds before harness attempts to close current test.
        timeout: Time in seconds before iteration is considered a timeout.
        no_harness: Indicate whether harness will is disabled.

    Returns:
        None
    """
    if timeout > 0:
        if no_harness:
            LOG.info("Using timeout: %ds, harness: DISABLED", timeout)
        else:
            LOG.info("Using time limit: %ds, timeout: %ds", time_limit, timeout)
            if time_limit == timeout:
                LOG.info("To avoid unnecessary relaunches set timeout > time limit")
    else:
        if no_harness:
            LOG.info("Using timeout: DISABLED, harness: DISABLED")
        else:
            LOG.info("Using time limit: %ds, timeout: DISABLED,", time_limit)
        LOG.warning("TIMEOUT DISABLED, not recommended for automation")


def get_certs() -> CertificateBundle:
    """Load or create a CertificateBundle. This will search for a cached
    certificate bundle before generating and caching a new bundle.
    The cache allows reuse across processes.

    Args:
        None.

    Returns:
        CertificateBundle
    """
    cached = find_cached("crypto")
    # create certificates if required
    if cached is None or not (cached / "certs").exists():
        with TemporaryDirectory() as tmp_path:
            certs = Path(tmp_path) / "certs"
            certs.mkdir(parents=True)
            CertificateBundle.create(certs)
            # add newly generated certs to cache
            cached = add_cached("crypto", certs)
    # copy data from cache
    path = grz_tmp("certs") / str(getpid())
    copytree(cached / "certs", path)
    return CertificateBundle.load(path)


def time_limits(
    time_limit: int | None,
    timeout: int | None,
    tests: Iterable[TestCase] | None = None,
    default_limit: int = DEFAULT_TIME_LIMIT,
    timeout_delay: int = TIMEOUT_DELAY,
) -> tuple[int, int]:
    """Determine the test time limit and timeout. If time_limit or timeout is None
    it is calculated otherwise the provided value is used.

    Args:
        time_limit: Test time limit.
        timeout: Iteration timeout.
        tests: Testcases that may contain time limit values.
        default_limit: Value to use as default time limit.
        timeout_delay: Value to use as delay when calculating timeout.

    Returns:
        Time limit and timeout.
    """
    assert default_limit > 0
    assert timeout_delay >= 0
    # calculate time limit
    calc_limit = time_limit is None
    if calc_limit:
        # use default_limit as a minimum
        test_limits = [default_limit]
        if tests:
            # add small time buffer to duration
            test_limits.extend(int(x.duration) + 10 for x in tests if x.duration)
        time_limit = max(test_limits)
    assert time_limit is not None and time_limit > 0
    # calculate timeout
    if timeout is None:
        timeout = time_limit + timeout_delay
    elif calc_limit and time_limit > timeout > 0:
        LOG.debug("calculated time limit > given timeout, using timeout")
        time_limit = timeout
    assert timeout >= 0
    # timeout should always be >= time limit unless timeout is disabled
    assert timeout >= time_limit or timeout == 0
    return time_limit, timeout



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""
unit tests for grizzly.replay
"""
from itertools import cycle
from pathlib import Path

from FTB.Signatures.CrashInfo import CrashSignature
from pytest import mark, raises

from sapphire import Served

from ..common.report import Report
from ..common.storage import TestCase, TestCaseLoadFailure
from ..target import AssetManager, Result, Target
from .replay import ReplayManager, ReplayResult

pytestmark = mark.usefixtures("patch_collector", "tmp_path_grz_tmp")


def _fake_save_logs(result_logs):
    """write fake log data to disk"""
    log_path = Path(result_logs)
    (log_path / "log_stderr.txt").write_text("STDERR log\n")
    (log_path / "log_stdout.txt").write_text("STDOUT log\n")
    with (log_path / "log_asan_blah.txt").open("w") as log_fp:
        log_fp.write("==1==ERROR: AddressSanitizer: ")
        log_fp.write("SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
        log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write("    #1 0x1337dd in bar /file2.c:1806:19\n")


@mark.parametrize("post_launch_delay", [0, -1])
def test_replay_01(mocker, server, tmp_path, post_launch_delay):
    """test ReplayManager.run() - no repro"""
    target = mocker.Mock(spec_set=Target, closed=True, launch_timeout=30)
    target.check_result.return_value = Result.NONE
    target.monitor.is_healthy.return_value = False
    (tmp_path / "a.html").touch()
    server.serve_path.return_value = (Served.ALL, {"a.html": str(tmp_path / "a.html")})
    with (
        TestCase.load(tmp_path) as testcase,
        ReplayManager([], server, target, use_harness=True, relaunch=1) as replay,
    ):
        assert not replay.run([testcase], 10, post_launch_delay=post_launch_delay)
        assert replay.signature is None
        assert replay.status
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results.total == 0
        assert target.monitor.is_healthy.call_count == 1
        assert target.close.call_count == 2
        assert target.close.mock_calls[0] == mocker.call()
        assert target.close.mock_calls[1] == mocker.call(force_close=True)


def test_replay_02(mocker, server, tmp_path):
    """test ReplayManager.run() - no repro - with repeats"""
    target = mocker.Mock(
        spec_set=Target, binary=Path("bin"), closed=False, launch_timeout=30
    )
    target.check_result.return_value = Result.NONE
    target.monitor.is_healthy.return_value = False
    iter_cb = mocker.Mock()
    (tmp_path / "a.html").touch()
    server.serve_path.return_value = (Served.ALL, {"a.html": str(tmp_path / "a.html")})
    with (
        TestCase.load(tmp_path) as testcase,
        ReplayManager([], server, target, use_harness=True, relaunch=20) as replay,
    ):
        assert not replay.run([testcase], 10, repeat=10, on_iteration_cb=iter_cb)
        assert replay.signature is None
        assert replay.status
        assert replay.status.ignored == 0
        assert replay.status.iteration == iter_cb.call_count == 10
        assert replay.status.results.total == 0
        assert target.handle_hang.call_count == 0
        assert target.monitor.is_healthy.call_count == 1
        assert target.close.call_count == 2


def test_replay_03(mocker, server, tmp_path):
    """test ReplayManager.run() - exit - skip shutdown in runner"""
    # this will make runner appear to have just relaunched the target
    # and skip the expected shutdown
    mocker.patch(
        "grizzly.common.runner.Runner._tests_run",
        new_callable=mocker.PropertyMock,
        return_value=0,
    )
    target = mocker.Mock(spec_set=Target, closed=False, launch_timeout=30)
    target.check_result.return_value = Result.NONE
    (tmp_path / "a.html").touch()
    server.serve_path.return_value = (Served.ALL, {"a.html": str(tmp_path / "a.html")})
    with (
        TestCase.load(tmp_path) as testcase,
        ReplayManager([], server, target, use_harness=True, relaunch=20) as replay,
    ):
        assert not replay.run([testcase], 10, repeat=10, min_results=1)
        assert replay.status
        assert replay.status.ignored == 0
        assert replay.status.iteration == 10
        assert replay.status.results.total == 0
        assert target.handle_hang.call_count == 0
        assert target.monitor.is_healthy.call_count == 0
        assert target.close.call_count == 1


@mark.parametrize("sig_parsed", [True, False])
@mark.parametrize("post_launch_delay", [0, -1])
def test_replay_04(mocker, server, tmp_path, sig_parsed, post_launch_delay):
    """test ReplayManager.run() - successful repro"""
    target = mocker.Mock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.monitor.is_healthy.return_value = False
    if sig_parsed:
        target.save_logs = _fake_save_logs
    else:

        def _save_logs(result_logs):
            """create uninteresting logs"""
            log_path = Path(result_logs)
            (log_path / "log_stderr.txt").write_text("STDERR log\n")
            (log_path / "log_stdout.txt").write_text("STDOUT log\n")

        target.save_logs = _save_logs
    (tmp_path / "a.html").touch()
    server.serve_path.return_value = (Served.ALL, {"a.html": str(tmp_path / "a.html")})
    with TestCase.load(tmp_path) as testcase:
        with ReplayManager([], server, target, relaunch=10) as replay:
            assert replay.signature is None
            results = replay.run([testcase], 10, post_launch_delay=post_launch_delay)
            if sig_parsed:
                assert replay.signature is not None
            else:
                assert replay.signature is None
            assert replay.status
            assert replay.status.ignored == 0
            assert replay.status.iteration == 1
            assert replay.status.results.total == 1
            assert target.handle_hang.call_count == 0
            assert target.monitor.is_healthy.call_count == 1
            assert target.close.call_count == 2
        assert len(results) == 1
        assert results[0].count == 1
        assert results[0].expected
        assert results[0].report
        assert len(results[0].durations) == 1
        results[0].report.cleanup()


def test_replay_05(mocker, server):
    """test ReplayManager.run() - error - entry point not requested"""
    target = mocker.Mock(
        spec_set=Target, binary=Path("bin"), closed=True, launch_timeout=30
    )
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html", assets={})]
    # test target unresponsive
    target.check_result.return_value = Result.NONE
    server.serve_path.return_value = (Served.NONE, {})
    with ReplayManager([], server, target, use_harness=False) as replay:
        assert not replay.run(tests, 10, repeat=1)
        assert replay.status
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results.total == 0
        # target.close() called once in runner and once by ReplayManager.run()
        assert target.close.call_count == 2
    target.reset_mock()


def test_replay_06(mocker, server):
    """test ReplayManager.run()
    delayed failure - following test entry point not requested"""
    target = mocker.Mock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    type(target).closed = mocker.PropertyMock(side_effect=(True, False, True))
    target.check_result.side_effect = (Result.NONE, Result.FOUND)
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    server.serve_path.side_effect = (
        (Served.ALL, {"a.html": "/fake/path"}),
        (Served.REQUEST, {"x": "/fake/path"}),
    )
    with ReplayManager([], server, target, use_harness=True, relaunch=10) as replay:
        assert replay.run(tests, 10, repeat=2)
        assert replay.status
        assert replay.status.ignored == 0
        assert replay.status.iteration == 2
        assert replay.status.results.total == 1
        # target.close() called once in runner and once by ReplayManager.run()
        assert target.close.call_count == 2


def test_replay_07(mocker, server):
    """test ReplayManager.run() - ignored (timeout)"""
    server.serve_path.return_value = (Served.TIMEOUT, {"a.html": "/fake/path"})
    target = mocker.Mock(spec_set=Target, closed=True, launch_timeout=30)
    target.check_result.return_value = Result.IGNORED
    target.handle_hang.return_value = True
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    with ReplayManager([], server, target, use_harness=False) as replay:
        assert not replay.run(tests, 10)
        assert replay.status
        assert replay.status.ignored == 1
        assert replay.status.iteration == 1
        assert replay.status.results.total == 0
    assert target.handle_hang.call_count == 1


def test_replay_08(mocker, server):
    """test ReplayManager.run() - early exit"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    server.serve_path.return_value = (Served.ALL, {"a.html": "/fake/path"})
    target = mocker.Mock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    target.save_logs = _fake_save_logs
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    # early failure
    target.check_result.side_effect = (
        Result.FOUND,
        Result.IGNORED,
        Result.NONE,
    )
    target.monitor.is_healthy.side_effect = (False, False, True, False)
    with ReplayManager([], server, target, use_harness=False) as replay:
        assert not replay.run(tests, 10, repeat=4, min_results=3)
        assert target.close.call_count == 4
        assert replay.status
        assert replay.status.iteration == 3
        assert replay.status.results.total == 1
        assert replay.status.ignored == 1
    # early success
    target.reset_mock()
    target.check_result.side_effect = (
        Result.FOUND,
        Result.IGNORED,
        Result.FOUND,
    )
    target.monitor.is_healthy.side_effect = (False, False, False)
    with ReplayManager([], server, target, use_harness=False) as replay:
        results = replay.run(tests, 10, repeat=4, min_results=2)
        assert target.close.call_count == 4
        assert replay.status
        assert replay.status.iteration == 3
        assert replay.status.results.total == 2
        assert replay.status.ignored == 1
    assert len(results) == 1
    assert sum(x.count for x in results) == 2
    target.reset_mock()
    # ignore early failure (perform all repeats)
    target.check_result.return_value = Result.NONE
    target.check_result.side_effect = None
    target.monitor.is_healthy.side_effect = None
    target.monitor.is_healthy.return_value = True
    with ReplayManager([], server, target, use_harness=False) as replay:
        assert not replay.run(tests, 10, repeat=4, min_results=4, exit_early=False)
        assert target.close.call_count == 5
        assert replay.status
        assert replay.status.iteration == 4
        assert replay.status.results.total == 0
        assert replay.status.ignored == 0
    target.reset_mock()
    # ignore early success (perform all repeats)
    target.check_result.return_value = Result.FOUND
    target.monitor.is_healthy.return_value = False
    with ReplayManager([], server, target, use_harness=False) as replay:
        results = replay.run(tests, 10, repeat=4, min_results=1, exit_early=False)
        assert target.close.call_count == 5
        assert replay.status
        assert replay.status.iteration == 4
        assert replay.status.results.total == 4
        assert replay.status.ignored == 0
    assert len(results) == 1
    assert sum(x.count for x in results) == 4


def test_replay_09(mocker, server):
    """test ReplayManager.run() - test signatures - fail to meet minimum"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    report_1 = mocker.Mock(spec_set=Report, crash_hash="h1", major="0123", minor="0123")
    report_2 = mocker.Mock(spec_set=Report, crash_hash="h2", major="0123", minor="abcd")
    report_3 = mocker.Mock(spec_set=Report, crash_hash="h2", major="0123", minor="abcd")
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_1, report_2, report_3)
    fake_report.calc_hash.return_value = "bucketHASH"
    server.serve_path.return_value = (Served.ALL, {"a.html": "/fake/path"})
    signature = mocker.Mock()
    signature.matches.side_effect = (True, False, False)
    target = mocker.Mock(spec_set=Target, launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    with ReplayManager(
        [], server, target, signature=signature, use_harness=False
    ) as replay:
        results = replay.run(tests, 10, repeat=3, min_results=2)
        assert target.close.call_count == 4
        assert replay.signature == signature
        assert replay.status
        assert replay.status.iteration == 3
        assert replay.status.results.total == 1
        assert replay.status.ignored == 2
    assert fake_report.call_count == 3
    assert len(results) == 1
    assert not results[0].expected
    assert results[0].count == 2
    assert report_1.cleanup.call_count == 1
    assert report_2.cleanup.call_count == 0
    assert report_3.cleanup.call_count == 1
    assert signature.matches.call_count == 3


def test_replay_10(mocker, server):
    """test ReplayManager.run() - test signatures - multiple matches"""
    report_0 = mocker.Mock(spec_set=Report, crash_hash="h1", major="0123", minor="0123")
    report_1 = mocker.Mock(spec_set=Report, crash_hash="h2", major="0123", minor="abcd")
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_0, report_1)
    fake_report.calc_hash.return_value = "bucketHASH"
    server.serve_path.return_value = (Served.ALL, {"a.html": "/fake/path"})
    sig = mocker.Mock(spec_set=CrashSignature)
    sig.matches.side_effect = (True, True)
    target = mocker.Mock(spec_set=Target, launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.monitor.is_healthy.return_value = False
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    with ReplayManager([], server, target, signature=sig, use_harness=False) as replay:
        results = replay.run(tests, 10, repeat=2, min_results=2)
        assert target.close.call_count == 3
        assert replay.signature == sig
        assert replay.status
        assert replay.status.iteration == 2
        assert replay.status.results.total == 2
        assert replay.status.ignored == 0
    assert fake_report.call_count == 2
    assert len(results) == 1
    assert results[0].expected
    assert results[0].count == 2
    assert report_0.cleanup.call_count == 0
    assert report_1.cleanup.call_count == 1
    assert sig.matches.call_count == 2


def test_replay_11(mocker, server):
    """test ReplayManager.run() - any crash - success"""
    report_1 = mocker.Mock(spec_set=Report, crash_hash="h1", major="0123", minor="0123")
    report_2 = mocker.Mock(spec_set=Report, crash_hash="h2", major="0123", minor="abcd")
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_1, report_2)
    server.serve_path.return_value = (Served.ALL, {"a.html": "/fake/path"})
    target = mocker.Mock(spec_set=Target, launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.monitor.is_healthy.return_value = False
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    with ReplayManager([], server, target, any_crash=True, use_harness=False) as replay:
        results = replay.run(tests, 10, repeat=2, min_results=2)
        assert target.close.call_count == 3
        assert replay.signature is None
        assert replay.status
        assert replay.status.iteration == 2
        assert replay.status.results.total == 2
        assert replay.status.ignored == 0
    assert fake_report.call_count == 2
    assert len(results) == 2
    assert all(x.expected for x in results)
    assert sum(x.count for x in results if x.expected) == 2
    assert report_1.cleanup.call_count == 0
    assert report_2.cleanup.call_count == 0


def test_replay_12(mocker, server):
    """test ReplayManager.run() - any crash - fail to meet minimum"""
    report_1 = mocker.Mock(spec_set=Report, crash_hash="h1", major="0123", minor="0123")
    report_2 = mocker.Mock(spec_set=Report, crash_hash="h2", major="0123", minor="abcd")
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_1, report_2)
    server.serve_path.return_value = (Served.ALL, {"a.html": "/fake/path"})
    target = mocker.Mock(spec_set=Target, launch_timeout=30)
    target.check_result.side_effect = (
        Result.NONE,
        Result.FOUND,
        Result.FOUND,
        Result.NONE,
    )
    target.monitor.is_healthy.return_value = False
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    with ReplayManager([], server, target, any_crash=True) as replay:
        assert not replay.run(tests, 10, repeat=4, min_results=3)
        assert target.close.call_count == 5
        assert replay.signature is None
        assert replay.status
        assert replay.status.iteration == 4
        assert replay.status.results.total == 2
        assert replay.status.ignored == 0
    assert fake_report.call_count == 2
    assert report_1.cleanup.call_count == 1
    assert report_2.cleanup.call_count == 1


@mark.parametrize(
    "to_serve, sig_value, expected, unexpected",
    [
        # No signature provided
        (((Served.NONE, {}), (Served.ALL, {"a.html": "/fake/path"})), None, 2, 0),
        (((Served.ALL, {"a.html": "/fake/path"}), (Served.NONE, {})), None, 2, 0),
        (((Served.NONE, {}), (Served.NONE, {})), None, 2, 0),
        # Signature provided (signatures match)
        (((Served.NONE, {}), (Served.ALL, {"a.html": "/fake/path"})), "STDERR", 2, 0),
        (((Served.ALL, {"a.html": "/fake/path"}), (Served.NONE, {})), "STDERR", 2, 0),
        (((Served.NONE, {}), (Served.NONE, {})), "STDERR", 2, 0),
        # Signature provided (signatures don't match)
        (((Served.NONE, {}), (Served.ALL, {"a.html": "/fake/path"})), "miss", 0, 1),
        (((Served.ALL, {"a.html": "/fake/path"}), (Served.NONE, {})), "miss", 0, 1),
        (((Served.NONE, {}), (Served.NONE, {})), "miss", 0, 1),
        # Hang after launch, before first test
        (((Served.TIMEOUT, {}),), None, 0, 0),
    ],
)
def test_replay_13(mocker, server, tmp_path, to_serve, sig_value, expected, unexpected):
    """test ReplayManager.run() - result after launch, before running test"""
    server.serve_path.side_effect = to_serve

    # prepare signature
    if sig_value is not None:
        sig_file = tmp_path / "sig.json"
        sig_file.write_text(
            "{\n"
            '  "symptoms": [\n'
            "    {\n"
            '      "src": "stderr",\n'
            '      "type": "output",\n'
            f'      "value": "/{sig_value}/"\n'
            "    }\n"
            "  ]\n"
            "}\n"
        )
        sig = CrashSignature.fromFile(str(sig_file))
    else:
        sig = None

    target = mocker.Mock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.save_logs = _fake_save_logs
    target.monitor.is_healthy.return_value = False
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    with ReplayManager(set(), server, target, any_crash=False, signature=sig) as replay:
        assert replay._relaunch == 1, "test is broken!"
        results = replay.run(tests, 10, repeat=2, min_results=2)
        assert sum(x.count for x in results if x.expected) == expected
        assert sum(x.count for x in results if not x.expected) == unexpected
        # make sure the serve flag is updated
        assert any(x.report.unstable for x in results) or expected == 0


def test_replay_14(mocker, server):
    """test ReplayManager.run() - no signature - use first crash"""
    auto_sig = mocker.Mock(spec_set=CrashSignature)
    auto_sig.matches.side_effect = (True, False, True)
    # original
    report_1 = mocker.Mock(spec_set=Report, crash_hash="h1", major="012", minor="999")
    report_1.crash_signature = auto_sig
    # non matching report
    report_2 = mocker.Mock(spec_set=Report, crash_hash="h2", major="abc", minor="987")
    # matching report
    report_3 = mocker.Mock(spec_set=Report, crash_hash="h1", major="012", minor="999")
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_1, report_2, report_3)
    fake_report.calc_hash.return_value = "bucket_hash"
    server.serve_path.return_value = (Served.ALL, {"a.html": "/fake/path"})
    target = mocker.Mock(spec_set=Target, launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.monitor.is_healthy.return_value = False
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    with ReplayManager([], server, target, use_harness=False) as replay:
        results = replay.run(tests, 10, repeat=3, min_results=2)
        assert target.close.call_count == 4
        assert replay.signature == auto_sig
        assert replay.status
        assert replay.status.iteration == 3
        assert replay.status.results.total == 2
        assert replay.status.ignored == 1
    assert fake_report.call_count == 3
    assert len(results) == 2
    assert sum(x.expected for x in results) == 1
    assert sum(x.count for x in results if x.expected) == 2
    assert report_1.cleanup.call_count == 0
    assert report_2.cleanup.call_count == 0
    assert report_3.cleanup.call_count == 1


def test_replay_15(mocker, server):
    """test ReplayManager.run() - unexpected exception"""
    report_0 = mocker.Mock(spec_set=Report, crash_hash="h1", major="0123", minor="0123")
    fake_report = mocker.patch("grizzly.replay.replay.Report", autospec=True)
    fake_report.side_effect = (report_0,)
    server.serve_path.side_effect = (
        (Served.ALL, {"a.html": "/fake/path"}),
        KeyboardInterrupt,
    )
    target = mocker.Mock(spec_set=Target, launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    tests = [mocker.MagicMock(spec_set=TestCase, entry_point="a.html")]
    with ReplayManager(
        [], server, target, any_crash=True, use_harness=True, relaunch=2
    ) as replay:
        with raises(KeyboardInterrupt):
            replay.run(tests, 10, repeat=3, min_results=2)
        assert replay.signature is None
        assert replay.status
        assert replay.status.iteration == 2
        assert replay.status.results.total == 1
        assert replay.status.ignored == 0
    assert target.close.call_count == 1
    assert target.monitor.is_healthy.call_count == 0
    assert fake_report.call_count == 1
    assert report_0.cleanup.call_count == 1


def test_replay_16(mocker, server):
    """test ReplayManager.run() - multiple TestCases - no repro"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    server.serve_path.return_value = (Served.ALL, {"a.html": "/fake/path"})
    target = mocker.Mock(spec_set=Target, closed=True, launch_timeout=30)
    target.check_result.return_value = Result.NONE
    tests = [
        mocker.MagicMock(spec_set=TestCase, entry_point="a.html") for _ in range(3)
    ]
    with ReplayManager([], server, target, use_harness=True) as replay:
        assert not replay.run(tests, 10)
        assert replay.status
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results.total == 0
    assert target.close.call_count == 2


def test_replay_17(mocker, server):
    """test ReplayManager.run() - multiple TestCases - no repro - with repeats"""
    server.serve_path.return_value = (Served.ALL, {"a.html": "/fake/path"})
    target = mocker.Mock(spec_set=Target, launch_timeout=30)
    # test relaunch < repeat
    type(target).closed = mocker.PropertyMock(side_effect=cycle([True, False]))
    target.check_result.return_value = Result.NONE
    target.monitor.is_healthy.return_value = False
    tests = [
        mocker.MagicMock(spec_set=TestCase, entry_point="a.html") for _ in range(3)
    ]
    with ReplayManager([], server, target, use_harness=True, relaunch=2) as replay:
        assert not replay.run(tests, 10, repeat=10)
        assert server.serve_path.call_count == 30
        assert target.close.call_count == 6
        assert target.launch.call_count == 5
        assert replay.status
        assert replay.status.ignored == 0
        assert replay.status.iteration == 10
        assert replay.status.results.total == 0
    assert target.monitor.is_healthy.call_count == 5


def test_replay_18(mocker, server):
    """test ReplayManager.run() - multiple TestCases - successful repro"""
    server.serve_path.side_effect = (
        (Served.ALL, {"0.html": "/fake/path"}),
        (Served.ALL, {"1.html": "/fake/path"}),
        (Served.ALL, {"2.html": "/fake/path"}),
    )
    target = mocker.Mock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    target.check_result.side_effect = (
        Result.NONE,
        Result.NONE,
        Result.FOUND,
    )
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    tests = [
        mocker.MagicMock(spec_set=TestCase, entry_point=f"{i}.html") for i in range(3)
    ]
    with ReplayManager([], server, target, use_harness=True) as replay:
        results = replay.run(tests, 30)
        assert target.close.call_count == 2
        assert replay.status
        assert replay.status.ignored == 0
        assert replay.status.iteration == 1
        assert replay.status.results.total == 1
    assert len(results) == 1
    assert len(results[0].durations) == len(tests)
    assert server.serve_path.call_count == 3


def test_replay_19(mocker, server, tmp_path):
    """test ReplayManager.run() - multiple calls"""
    mocker.patch("grizzly.common.runner.sleep", autospec=True)
    target = mocker.Mock(spec_set=Target, closed=True, launch_timeout=30)
    target.check_result.return_value = Result.NONE
    (tmp_path / "test.html").touch()
    server.serve_path.return_value = (
        Served.ALL,
        {"test.html": str(tmp_path / "test.html")},
    )
    with (
        TestCase.load(tmp_path) as testcase,
        ReplayManager([], server, target, use_harness=True) as replay,
    ):
        assert not replay.run([testcase], 30)
        assert replay.status
        assert replay.status.iteration == 1
        assert not replay.run([testcase], 30)
        assert replay.status
        assert replay.status.iteration == 1
        assert not replay.run([testcase], 30)
        assert replay.status
        assert replay.status.iteration == 1
    assert server.serve_path.call_count == 3


def test_replay_20(mocker, tmp_path):
    """test ReplayManager.report_to_filesystem()"""
    # no reports
    ReplayManager.report_to_filesystem(tmp_path, [], [])
    assert not any(tmp_path.iterdir())
    # with reports and tests
    (tmp_path / "expected_a").mkdir()
    result0 = ReplayResult(
        mocker.Mock(spec_set=Report, path=tmp_path / "expected_a", prefix="expected_1"),
        [1],
        True,
    )
    (tmp_path / "other_a").mkdir()
    result1 = ReplayResult(
        mocker.Mock(spec_set=Report, path=tmp_path / "other_a", prefix="other_1"),
        [1],
        False,
    )
    (tmp_path / "other_b").mkdir()
    result2 = ReplayResult(
        mocker.Mock(spec_set=Report, path=tmp_path / "other_b", prefix="other_2"),
        [1],
        False,
    )
    test = mocker.Mock(spec_set=TestCase, timestamp=1.0)
    path = tmp_path / "dest"
    ReplayManager.report_to_filesystem(path, [result0, result1, result2], [test])
    assert test.dump.call_count == 3  # called once per report
    assert path.is_dir()
    assert (path / "reports").is_dir()
    assert (path / "reports" / "expected_1_logs").is_dir()
    assert (path / "other_reports").is_dir()
    assert (path / "other_reports" / "other_1_logs").is_dir()
    assert (path / "other_reports" / "other_2_logs").is_dir()
    # with reports and no tests
    (tmp_path / "expected_b").mkdir()
    (tmp_path / "report_expected").mkdir()
    result3 = ReplayResult(
        mocker.Mock(spec_set=Report, path=tmp_path / "expected_b", prefix="expected_2"),
        [1],
        True,
    )
    path = tmp_path / "dest2"
    ReplayManager.report_to_filesystem(path, [result3], [])
    assert not (tmp_path / "expected_2").is_dir()
    assert path.is_dir()
    assert (path / "reports" / "expected_2_logs").is_dir()


def test_replay_21(tmp_path):
    """test ReplayManager.load_testcases()"""
    # nothing to load
    with raises(TestCaseLoadFailure, match="Failed to load TestCases"):
        ReplayManager.load_testcases([])
    input_path = tmp_path / "test.html"
    input_path.touch()
    # load single file
    tests, asset_mgr, env_vars = ReplayManager.load_testcases([input_path])
    assert len(tests) == 1
    assert not tests[-1].assets
    assert not tests[-1].env_vars
    assert not env_vars
    assert asset_mgr is None
    # skip invalid
    tests, _, _ = ReplayManager.load_testcases([Path("missing"), input_path])
    assert len(tests) == 1


def test_replay_22(tmp_path):
    """test ReplayManager.load_testcases() with assets and env vars"""
    # build test case
    with AssetManager() as asset_mgr:
        (tmp_path / "prefs.js").touch()
        asset_mgr.add("prefs", tmp_path / "prefs.js", copy=False)
        with TestCase("test.html", "foo") as test:
            test.add_from_bytes(b"", test.entry_point)
            test.env_vars = {"foo": "bar"}
            test.assets = dict(asset_mgr.assets)
            test.assets_path = asset_mgr.path
            test.dump(tmp_path / "src", include_details=True)
    # load directory with test info file
    tests, asset_mgr, env_vars = ReplayManager.load_testcases([tmp_path / "src"])
    assert asset_mgr
    with asset_mgr:
        assert len(tests) == 1
        assert not tests[-1].assets
        assert not tests[-1].env_vars
        assert env_vars == {"foo": "bar"}
        assert "prefs" in asset_mgr.assets


@mark.parametrize(
    "expect_hang, is_hang, use_sig, match_sig, ignored, results",
    [
        # reproduce expected hang
        (True, True, True, True, 0, 1),
        # expected hang (signature, hang - no match)
        (True, True, True, False, 1, 0),
        # expected hang got crash (signature)
        (True, False, True, False, 1, 0),
        # unexpected hang (signature, no match)
        (False, True, True, False, 1, 0),
        # unexpected hang (no signature)
        (False, True, False, False, 1, 0),
        # unexpected crash (signature)
        (False, False, True, False, 1, 0),
    ],
)
def test_replay_23(
    mocker, server, expect_hang, is_hang, use_sig, match_sig, ignored, results
):
    """test ReplayManager.run() - detect hangs"""
    server.serve_path.return_value = (
        Served.TIMEOUT if is_hang else Served.ALL,
        {"a.html": "/fake/path"},
    )
    if use_sig:
        signature = mocker.Mock()
        signature.matches.return_value = match_sig
        signature.rawSignature = "fakesig"
    else:
        signature = None
    target = mocker.Mock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.handle_hang.return_value = False
    target.save_logs = _fake_save_logs
    target.monitor.is_healthy.return_value = False
    test = mocker.MagicMock(spec_set=TestCase, entry_point="a.html", hang=is_hang)
    with ReplayManager([], server, target, signature=signature, relaunch=10) as replay:
        found = replay.run([test], 10, expect_hang=expect_hang)
        assert replay.status
        assert replay.status.iteration == 1
        assert replay.status.ignored == ignored
        assert replay.status.results.total == results
        assert target.handle_hang.call_count == (1 if is_hang else 0)
        assert target.monitor.is_healthy.call_count == (0 if is_hang else 1)
        assert target.close.call_count == (1 if is_hang else 2)
    assert len(found) == 1
    assert found[0].count == 1
    assert found[0].expected == results
    assert found[0].report
    assert found[0].report.is_hang == is_hang
    assert len(found[0].durations) == 1
    found[0].report.cleanup()


def test_replay_24(mocker):
    """test ReplayManager.report_to_fuzzmanager()"""
    reporter = mocker.patch("grizzly.replay.replay.FuzzManagerReporter")
    # no reports or tests
    ReplayManager.report_to_fuzzmanager([], [], "grizzly-replay")
    assert reporter.call_args == (("grizzly-replay",),)
    assert reporter.return_value.submit.call_count == 0
    reporter.reset_mock()
    # with reports and tests
    ReplayManager.report_to_fuzzmanager(
        [
            ReplayResult(mocker.Mock(spec_set=Report), [3.1], True),
            ReplayResult(mocker.Mock(spec_set=Report), [10.2], False),
        ],
        [mocker.Mock(spec_set=TestCase)],
        "grizzly-replay",
    )
    assert reporter.return_value.submit.call_count == 2


def test_replay_25(mocker, server, tmp_path):
    """test ReplayManager.run() - signature - matching stacks"""
    sig_file = tmp_path / "sig.json"
    sig_file.write_text(
        "{\n"
        '  "symptoms": [\n'
        "    {\n"
        '      "src": "stderr",\n'
        '      "type": "output",\n'
        '      "value": "/STDERR/"\n'
        "    }\n"
        "  ]\n"
        "}\n"
    )
    sig = CrashSignature.fromFile(str(sig_file))

    target = mocker.MagicMock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.monitor.is_healthy.return_value = False

    call_count = 0

    def _save_logs_variation(result_logs):
        """create different stacks each call"""
        nonlocal call_count
        call_count += 1
        log_path = Path(result_logs)
        (log_path / "log_stderr.txt").write_text("STDERR log\n")
        (log_path / "log_stdout.txt").write_text("STDOUT log\n")
        with (log_path / "log_asan_blah.txt").open("w") as log_fp:
            log_fp.write("==1==ERROR: AddressSanitizer: ")
            log_fp.write("SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
            log_fp.write(f"    #0 0xbad000 in call_a{call_count:02d} file.c:23:34\n")
            log_fp.write(f"    #1 0xbad001 in call_b{call_count:02d} file.c:12:45\n")

    target.save_logs.side_effect = _save_logs_variation

    (tmp_path / "a.html").touch()
    server.serve_path.return_value = (Served.ALL, {"a.html": str(tmp_path / "a.html")})
    with TestCase.load(tmp_path) as testcase:
        with ReplayManager([], server, target, relaunch=10, signature=sig) as replay:
            results = replay.run([testcase], 10, min_results=2, repeat=2)
            assert replay.signature is not None
            assert replay.status
            assert replay.status.ignored == 0
            assert replay.status.iteration == 2
            assert replay.status.results.total == 2
        assert len(results) == 1
        assert results[0].count == 2
        assert results[0].expected
        assert results[0].report
        results[0].report.cleanup()


@mark.parametrize(
    "stderr_log, ignored, total, include_stack",
    [
        # match stack only
        (["STDERR log\n", "STDERR log\n"], 0, 2, [True] * 2),
        # match stack only
        (["STDERR log\n", "STDERR log\nAssertion failure: test\n"], 0, 2, [True] * 2),
        # match stack and assertion message
        (["STDERR log\nAssertion failure: test\n", "STDERR log\n"], 1, 1, [True] * 2),
        # match stack and assertion message
        (["Assertion failure: 1\n", "Assertion failure: 2\n"], 1, 1, [True] * 2),
        # match, no match, match
        (["Assertion failure: 1\n", "a\n", "Assertion failure: 1\n"], 1, 2, [True] * 3),
        # fail to create signature x2 (missing stack)
        (["STDERR log\n", "STDERR log\n"], 0, 2, [False] * 2),
        # fail to create signature, create signature
        (["STDERR log\n", "STDERR log\n"], 1, 1, [False, True]),
        # create signature, fail to create signature
        (["STDERR log\n", "STDERR log\n"], 1, 1, [True, False]),
    ],
)
def test_replay_26(mocker, server, tmp_path, stderr_log, ignored, total, include_stack):
    """test ReplayManager.run() - no signature - match first result"""
    # NOTE: this is similar to "no signature - use first crash" test
    # but this is more of an integration test
    iters = len(stderr_log)
    assert iters == ignored + total, "test is broken"
    assert iters == len(include_stack), "test is broken"
    target = mocker.Mock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    target.check_result.return_value = Result.FOUND
    target.monitor.is_healthy.return_value = False

    def _save_logs_variation(result_logs):
        """create logs"""
        nonlocal stderr_log
        nonlocal include_stack
        log_path = Path(result_logs)
        (log_path / "log_stderr.txt").write_text(stderr_log.pop(0))
        (log_path / "log_stdout.txt").write_text("STDOUT log\n")
        if include_stack.pop(0):
            with (log_path / "log_asan_blah.txt").open("w") as log_fp:
                log_fp.write("==1==ERROR: AddressSanitizer: ")
                log_fp.write("SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n")
                log_fp.write("    #0 0xbad000 in call_a file.c:23:34\n")
                log_fp.write("    #1 0xbad001 in call_b file.c:12:45\n")

    target.save_logs.side_effect = _save_logs_variation

    has_sig = include_stack[0]
    (tmp_path / "a.html").touch()
    server.serve_path.return_value = (Served.ALL, {"a.html": str(tmp_path / "a.html")})
    with TestCase.load(tmp_path) as testcase:
        with ReplayManager([], server, target, relaunch=10) as replay:
            results = replay.run([testcase], 10, min_results=2, repeat=iters)
            if has_sig:
                assert replay.signature is not None
            else:
                assert replay.signature is None
            assert replay.status
            assert replay.status.ignored == ignored
            assert replay.status.iteration == iters
            assert replay.status.results.total == total
        for result in results:
            result.report.cleanup()
        assert results


def test_replay_27(mocker, server, tmp_path):
    """test ReplayManager.run() - collect served files"""
    target = mocker.Mock(spec_set=Target, binary=Path("bin"), launch_timeout=30)
    target.check_result.side_effect = (Result.FOUND,)
    target.monitor.is_healthy.return_value = False
    target.save_logs = _fake_save_logs
    src = tmp_path / "src"
    with TestCase("test.html", "test-adapter") as test:
        test.add_from_bytes(b"", test.entry_point)
        test.add_from_bytes(b"", "include.js", required=False)
        test.add_from_bytes(b"", "no-serve.html", required=False)
        test.dump(src, include_details=True)
    with TestCase.load(src) as test:
        server.serve_path.side_effect = (
            (
                Served.ALL,
                {
                    "test.html": test.root / "test.html",
                    "include.js": test.root / "include.js",
                },
            ),
        )
        with ReplayManager([], server, target, use_harness=True) as replay:
            assert "include.js" not in test
            results = replay.run([test], 30)
            assert replay.status
            assert replay.status.ignored == 0
            assert replay.status.iteration == 1
            assert replay.status.results.total == 1
            assert len(results) == 1
            assert "test.html" in test
            assert "include.js" in test
            assert "no-serve.html" not in test
            replay.report_to_filesystem(tmp_path / "dst", results, [test])
    assert (tmp_path / "dst").rglob("test.html")
    assert (tmp_path / "dst").rglob("include.js")
    assert not any((tmp_path / "dst").rglob("no-serve.html"))


@mark.parametrize(
    "adapters, expected",
    [
        # no test cases
        ([], None),
        # empty adapter name
        ([""], None),
        # multiple test cases
        (["test-adapter", "", None, "test-adapter"], "grizzly-test-adapter"),
    ],
)
def test_replay_28(mocker, adapters, expected):
    """test ReplayManager.lookup_tool()"""
    testcases = [mocker.Mock(spec_set=TestCase, adapter_name=name) for name in adapters]
    assert ReplayManager.lookup_tool(testcases) == expected
